# pylint: disable=C0103,C0111,C0302,W0212

import pytest
import quasardb


def test_max_cardinality_ok(qdbd_connection):
    qdbd_connection.options().set_max_cardinality(140000)


def test_max_cardinality_throws_when_value_is_zero(qdbd_connection):
    with pytest.raises(quasardb.InvalidArgumentError):
        qdbd_connection.options().set_max_cardinality(0)


def test_max_cardinality_throws_when_value_is_negative(qdbd_connection):
    with pytest.raises(TypeError):
        qdbd_connection.options().set_max_cardinality(-143)



# pylint: disable=C0103,C0111,C0302,W0212

def test_empty_prefix(qdbd_connection):

    res = qdbd_connection.prefix_get("testazeazeaze", 10)
    assert len(res) == 0

    res = qdbd_connection.prefix_count("testazeazeaze")
    assert res == 0


def test_find_one(qdbd_connection, entry_name, random_string):
    dat_prefix = random_string

    entry_name = dat_prefix + entry_name
    entry_content = "content"

    b = qdbd_connection.blob(entry_name)
    b.put(entry_content)

    res = qdbd_connection.prefix_get(dat_prefix, 10)
    assert len(res) == 1
    assert res[0] == entry_name

    assert qdbd_connection.prefix_count(dat_prefix) == 1



# pylint: disable=C0103,C0111,C0302,W0212
from builtins import range as xrange, int as long  # pylint: disable=W0622
from functools import reduce  # pylint: disable=W0622
import datetime
import test_table as tslib
from time import sleep
import time
import random

import pytest
import quasardb
import numpy as np

table_count, column_count, row_factor = 10000, 6, 1000
row_count = table_count * row_factor


def make_tables(qdbd_connection):
    table_creation_start = time.time()
    tables = []
    for i in range(table_count):
        name = 'table_%s' % (i,)
        cols = [
            quasardb.ColumnInfo(
                quasardb.ColumnType.Int64,
                'col_{}'.format(col_idx)) for col_idx in range(column_count)]
        qdbd_connection.query("DROP TABLE IF EXISTS {}".format(name))
        t = qdbd_connection.table(name)
        t.create(columns=cols, shard_size=datetime.timedelta(seconds=60))
        t.attach_tag('test_tag')
        tables.append(t)
    table_creation_time = time.time() - table_creation_start
    return table_creation_time, tables


def make_batch_columns():
    batch_columns = [
        quasardb.BatchColumnInfo(
            'table_{}'.format(tbl_idx),
            'col_{}'.format(col_idx),
            1) for tbl_idx in range(table_count) for col_idx in range(column_count)]
    return batch_columns


@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
def test_batch_insert(qdbd_connection):
    table_creation_time, _ = make_tables(qdbd_connection)

    inserter_creation_start = time.time()
    batch_columns = make_batch_columns()
    inserter = qdbd_connection.inserter(batch_columns)
    inserter_creation_time = time.time() - inserter_creation_start

    print(f'{__name__}:')
    print(f'  - {table_count} table(s)')
    print(f'  - {column_count} column(s) per table')
    print(f'  - {row_count} row(s)')
    print(f'  - table creation:     {table_creation_time}s')
    print(f'  - inserters creation: {inserter_creation_time}s')

    total_insertion_start = time.time()
    bulk_writing_start = time.time()
    bulk_start_row_time = 0.0

    for row_index in range(row_count):
        ts = int(time.time() * 1.0e9)
        bulk_start_row_start = time.time()
        inserter.start_row(ts)
        bulk_start_row_time = bulk_start_row_time + \
            (time.time() - bulk_start_row_start)
        tbl_idx = row_index % table_count
        for col_idx in range(column_count):
            inserter.set_int64(tbl_idx * column_count + col_idx,
                               int(row_index + random.random()))

    bulk_writing_time = time.time() - bulk_writing_start
    bulk_insert_start = time.time()
    inserter.push()
    bulk_insert_time = time.time() - bulk_insert_start
    total_insertion_time = time.time() - total_insertion_start

    res = qdbd_connection.query(
        "SELECT count(col_0) FROM FIND(tag='test_tag')")
    print(f'Results:')
    print(f'  - batch start row:     {bulk_start_row_time}s')
    print(f'  - batch set values:    {bulk_writing_time}s')
    print(f'  - batch insert values: {bulk_insert_time}s')
    print(f'  - total insert time:   {total_insertion_time}s')
    print(f'  - rows inserted: {res[0]}')


@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
def test_pinned_writer(qdbd_connection):
    table_creation_time, tables = make_tables(qdbd_connection)

    inserter_creation_start = time.time()
    inserter = qdbd_connection.pinned_writer(tables)
    inserter_creation_time = time.time() - inserter_creation_start

    print(f'{__name__}:')
    print(f'  - {table_count} table(s)')
    print(f'  - {column_count} column(s) per table')
    print(f'  - {row_count} row(s)')
    print(f'  - table creation:     {table_creation_time}s')
    print(f'  - inserters creation: {inserter_creation_time}s')

    total_insertion_start = time.time()
    bulk_writing_start = time.time()
    bulk_start_row_time = 0.0

    for row_index in range(row_count):
        ts = int(time.time() * 1.0e9)
        bulk_start_row_start = time.time()
        inserter.start_row(ts)
        bulk_start_row_time = bulk_start_row_time + \
            (time.time() - bulk_start_row_start)
        tbl_idx = row_index % table_count
        for col_idx in range(column_count):
            inserter.set_int64(tbl_idx * column_count + col_idx,
                               int(row_index + random.random()))

    bulk_writing_time = time.time() - bulk_writing_start
    bulk_insert_start = time.time()
    inserter.push()
    bulk_insert_time = time.time() - bulk_insert_start
    total_insertion_time = time.time() - total_insertion_start

    res = qdbd_connection.query(
        "SELECT count(col_0) FROM FIND(tag='test_tag')")
    print(f'Results:')
    print(f'  - batch start row:     {bulk_start_row_time}s')
    print(f'  - batch set values:    {bulk_writing_time}s')
    print(f'  - batch insert values: {bulk_insert_time}s')
    print(f'  - total insert time:   {total_insertion_time}s')
    print(f'  - rows inserted: {res[0]}')


@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
def test_pinned_writer_column(qdbd_connection):
    table_creation_time, tables = make_tables(qdbd_connection)

    inserter_creation_start = time.time()
    inserter = qdbd_connection.pinned_writer(tables)
    inserter_creation_time = time.time() - inserter_creation_start

    print(f'{__name__}:')
    print(f'  - {table_count} table(s)')
    print(f'  - {column_count} column(s) per table')
    print(f'  - {row_count} row(s)')
    print(f'  - table creation:     {table_creation_time}s')
    print(f'  - inserters creation: {inserter_creation_time}s')

    total_insertion_start = time.time()
    bulk_writing_start = time.time()
    bulk_start_row_time = 0.0
    timestamps = []
    values = []
    for tbl_idx in range(table_count):
        timestamps.append([])
        values.append([])
        for col_idx in range(column_count):
            timestamps[tbl_idx].append([])
            values[tbl_idx].append([])

    for row_idx in range(row_count):
        bulk_start_row_start = time.time()
        bulk_start_row_time = bulk_start_row_time + \
            (time.time() - bulk_start_row_start)
        tbl_idx = row_idx % table_count
        for col_idx in range(column_count):
            timestamps[tbl_idx][col_idx].append(int(time.time() * 1.0e9))
            values[tbl_idx][col_idx].append(int(row_idx + random.random()))

    for tbl_idx in range(table_count):
        for col_idx in range(column_count):
            inserter.set_int64_column(
                tbl_idx * column_count + col_idx,
                timestamps[tbl_idx][col_idx],
                values[tbl_idx][col_idx])

    bulk_writing_time = time.time() - bulk_writing_start
    bulk_insert_start = time.time()
    inserter.push()
    bulk_insert_time = time.time() - bulk_insert_start
    total_insertion_time = time.time() - total_insertion_start

    res = qdbd_connection.query(
        "SELECT count(col_0) FROM FIND(tag='test_tag')")
    print(f'Results:')
    print(f'  - batch start row:     {bulk_start_row_time}s')
    print(f'  - batch set values:    {bulk_writing_time}s')
    print(f'  - batch insert values: {bulk_insert_time}s')
    print(f'  - total insert time:   {total_insertion_time}s')
    print(f'  - rows inserted: {res[0]}')



# pylint: disable=C0103,C0111,C0302,W0212
from builtins import range as xrange  # pylint: disable=W0622
import pytest
import quasardb


def test_put(double_entry, random_double):
    double_entry.put(random_double)


def test_get(double_entry, random_double):
    double_entry.put(random_double)
    got = double_entry.get()

    assert random_double == got


def test_update(double_entry, random_double):
    double_entry.put(random_double)
    got = double_entry.get()

    assert random_double == got

    new_double = 42
    double_entry.update(new_double)
    got = double_entry.get()

    assert 42 == got


def test_add(double_entry, random_double):
    double_entry.put(random_double)
    got = double_entry.get()

    assert random_double == got

    double_entry.add(42)
    got = double_entry.get()

    assert (random_double + 42) == got



# pylint: disable=C0103,C0111,C0302,W0212
import numpy as np
import test_batch_inserter as batchlib
import time
import multiprocessing
from sys import platform


def _ensure_timeout(f, timeout=3):
    # HACKS(leon): apparently there is no reliable way to do this kind of stuff
    # on Windows or OSX, due to the way threading works.
    #
    # For now, this check always returns true on non-Linux OS'es, so that at least
    # we do have this check somewhere.
    #
    # The problem is inherently related to Python's single threaded-ness:
    # - if we use a thread-based approach (such as with python's `stopit` library),
    #   it also invalidates the cursor/iterator because we interrupt it.
    # - signals are not available under non-Unix OS'es, so we cannot us ethat either
    #   (and again, we would have to interrupt the next() function call somehow)
    # - if we use the multiprocess as defined below, it cannot be pickled correctly
    #   under any non-Linux OS.
    #
    # This test is useful, so if you know how to fix this under e.g. Windows, let's
    # fix it!
    if platform == "linux" or platform == "linux2":
        p = multiprocessing.Process(target=f)
        p.start()

        p.join(timeout)

        if not p.is_alive():
            return False

        p.terminate()
        p.join()
        return True

    else:
        # Non-linux OS
        return True


def test_subscribe_single_table(qdbd_connection, table, many_intervals):
    # Our test case for streaming data is as follows:
    # 1. We first initialize a subscription
    # 2. *After* this, we insert new data
    # 3. We ensure we actually retrieved that data, and exactly that data (nothing more)
    # 4. Insert new data
    # 5. Repeat step 3 again

    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    xs = table.subscribe(qdbd_connection)

    doubles, blobs, strings, integers, timestamps, symbols = batchlib._test_with_table(
        inserter, table, many_intervals, batchlib._regular_push)

    time.sleep(4)

    n = len(doubles)

    # After insertion, we expect exactly `n` items in the database, and not a
    # single element more.
    offset = 0
    while offset < n:
        row = next(xs)
        assert row['the_double'] == doubles[offset]
        assert row['the_blob'] == blobs[offset]
        assert row['the_string'] == strings[offset]
        assert row['the_int64'] == integers[offset]
        assert row['the_ts'] == timestamps[offset]
        assert row['the_symbol'] == symbols[offset]
        offset = offset + 1

    def next_row():
        return next(xs)

    # Ensure that acquiring the next row times out after 3 seconds, which ensures
    # that we have reached 'the end'.
    assert _ensure_timeout(next_row, timeout=3)

    # Now, we insert additional data, which should wake up our subscription.
    # What we're going to do is just insert the exact same data, but this time
    # with different intervals
    many_intervals_ = list()
    for x in many_intervals:
        many_intervals_.append(x + np.timedelta64(365, 'D'))

    doubles, blobs, strings, integers, timestamps, symbols = batchlib._test_with_table(
        inserter, table, many_intervals_, batchlib._regular_push)

    # Note that we only reset `offset`; since we just inserted exactly double
    # the data, `n` is still 10000 which is perfect!
    offset = 0
    while offset < n:
        row = next(xs)
        assert row['the_double'] == doubles[offset]
        assert row['the_blob'] == blobs[offset]
        assert row['the_string'] == strings[offset]
        assert row['the_int64'] == integers[offset]
        assert row['the_ts'] == timestamps[offset]
        assert row['the_symbol'] == symbols[offset]
        offset = offset + 1

    # Ensure that acquiring the next row times out after 3 seconds, which ensures
    # that we have reached 'the end'.
    assert _ensure_timeout(next_row, timeout=3)



# pylint: disable=C0103,C0111,C0302,W0212
from builtins import range as xrange  # pylint: disable=W0622
import pytest
import quasardb


def test_put(string_entry, random_string):
    string_entry.put(random_string)


def test_put_throws_exception_when_called_twice(string_entry, random_string):
    string_entry.put(random_string)
    with pytest.raises(quasardb.AliasAlreadyExistsError):
        string_entry.put(random_string)


def test_get(string_entry, random_string):
    string_entry.put(random_string)

    got = string_entry.get()

    assert random_string == got


def test_remove(string_entry, random_string):
    string_entry.put(random_string)
    string_entry.remove()

    with pytest.raises(quasardb.AliasNotFoundError):
        string_entry.get()


def test_remove_throws_exception_when_called_twice(string_entry, random_string):
    string_entry.put(random_string)
    string_entry.remove()

    with pytest.raises(quasardb.AliasNotFoundError):
        string_entry.remove()




def test_update(string_entry, random_string):
    string_entry.update(random_string)
    got = string_entry.get()
    assert random_string == got

    new_entry_content = "It's the new style"
    string_entry.update(new_entry_content)
    got = string_entry.get()
    assert got == new_entry_content

    string_entry.remove()

    new_entry_content = "I'm never in training, my voice is not straining"
    string_entry.update(new_entry_content)
    got = string_entry.get()

    assert got == new_entry_content


def test_get_and_update(string_entry, random_string):
    string_entry.put(random_string)
    got = string_entry.get()
    assert random_string == got

    entry_new_content = "new stuff"
    string_entry.update(entry_new_content)
    got = string_entry.get()

    assert entry_new_content == got

    got = string_entry.get()
    assert entry_new_content == got
    string_entry.remove()


def test_get_and_remove(string_entry, random_string):
    string_entry.put(random_string)

    got = string_entry.get_and_remove()
    assert random_string == got

    with pytest.raises(quasardb.AliasNotFoundError):
        string_entry.get()


def test_remove_if(string_entry, random_string):
    string_entry.put(random_string)
    got = string_entry.get()

    assert random_string == got

    with pytest.raises(quasardb.Error):
        string_entry.remove_if(random_string + "a")

    got = string_entry.get()
    assert random_string == got

    string_entry.remove_if(random_string)

    with pytest.raises(quasardb.AliasNotFoundError):
        string_entry.get()


def test_compare_and_swap(string_entry, random_string):

    string_entry.put(random_string)
    got = string_entry.get()

    assert random_string == got

    entry_new_content = "new stuff"
    got = string_entry.compare_and_swap(entry_new_content, entry_new_content)

    assert random_string == got

    got = string_entry.compare_and_swap(entry_new_content, random_string)
    assert len(got) == 0

    got = string_entry.get()
    assert entry_new_content == got

    string_entry.remove()



import pytest
import quasardb


def test_node_wrong_leading_qdb_in_uri(qdbd_connection, qdbd_settings):
    with pytest.raises(quasardb.Error, match=r'The (handle|argument) is invalid'):
        # With check in connect, node() throws.
        direct_conn = qdbd_connection.node(
            qdbd_settings.get("uri").get("insecure"))
        # Without it, C API should return an error in prefix_get.
        direct_conn.prefix_get("$qdb", 10)


def test_node_connect_throws_connection_error_when_no_cluster_public_key(
        qdbd_settings):
    with pytest.raises(quasardb.Error):
        quasardb.Node(
            uri=qdbd_settings.get("uri").get("secure").replace("qdb://", ""),
            user_name=qdbd_settings.get("security").get("user_name"),
            user_private_key=qdbd_settings.get("security").get("user_private_key"))


def test_node_connect_throws_connection_error_when_no_user_name(qdbd_settings):
    with pytest.raises(quasardb.Error):
        quasardb.Node(
            uri=qdbd_settings.get("uri").get("secure").replace("qdb://", ""),
            user_private_key=qdbd_settings.get("security").get("user_private_key"),
            cluster_public_key=qdbd_settings.get("security").get("cluster_public_key"))


def test_node_connect_throws_connection_error_when_no_user_private_key(
        qdbd_settings):
    with pytest.raises(quasardb.Error):
        quasardb.Node(
            uri=qdbd_settings.get("uri").get("secure").replace("qdb://", ""),
            user_name=qdbd_settings.get("security").get("user_name"),
            cluster_public_key=qdbd_settings.get("security").get("cluster_public_key"))


def test_node_connect_ok_to_secure_cluster(qdbd_settings):
    quasardb.Node(
        uri=qdbd_settings.get("uri").get("secure").replace("qdb://", ""),
        user_name=qdbd_settings.get("security").get("user_name"),
        user_private_key=qdbd_settings.get("security").get("user_private_key"),
        cluster_public_key=qdbd_settings.get("security").get("cluster_public_key"))


def test_node_connect_throws_connection_error_when_no_cluster_public_key_file(
        qdbd_settings):
    with pytest.raises(quasardb.Error):
        quasardb.Node(
            uri=qdbd_settings.get("uri").get("secure").replace("qdb://", ""),
            user_security_file=qdbd_settings.get("security").get("user_private_key_file"))


def test_node_connect_throws_connection_error_when_no_user_security_file(
        qdbd_settings):
    with pytest.raises(quasardb.Error):
        quasardb.Node(
            uri=qdbd_settings.get("uri").get("secure").replace("qdb://", ""),
            cluster_public_key_file=qdbd_settings.get("security").get("cluster_public_key_file"))


def test_node_connect_throws_connection_error_when_mix_security_1(
        qdbd_settings):
    with pytest.raises(quasardb.Error):
        quasardb.Node(
            uri=qdbd_settings.get("uri").get("secure").replace("qdb://", ""),
            user_name=qdbd_settings.get("security").get("user_name"),
            user_private_key=qdbd_settings.get("security").get("user_private_key"),
            cluster_public_key=qdbd_settings.get("security").get("cluster_public_key"),
            cluster_public_key_file=qdbd_settings.get("security").get("cluster_public_key_file"))


def test_node_connect_throws_connection_error_when_mix_security_2(
        qdbd_settings):
    with pytest.raises(quasardb.Error):
        quasardb.Node(
            uri=qdbd_settings.get("uri").get("secure").replace("qdb://", ""),
            user_name=qdbd_settings.get("security").get("user_name"),
            user_private_key=qdbd_settings.get("security").get("user_private_key"),
            user_security_file=qdbd_settings.get("security").get("user_private_key_file"),
            cluster_public_key_file=qdbd_settings.get("security").get("cluster_public_key_file"))


def test_node_connect_throws_connection_error_when_mix_security_full(
        qdbd_settings):
    with pytest.raises(quasardb.Error):
        quasardb.Node(
            uri=qdbd_settings.get("uri").get("secure").replace("qdb://", ""),
            user_name=qdbd_settings.get("security").get("user_name"),
            user_private_key=qdbd_settings.get("security").get("user_private_key"),
            cluster_public_key=qdbd_settings.get("security").get("cluster_public_key"),
            user_security_file=qdbd_settings.get("security").get("user_private_key_file"),
            cluster_public_key_file=qdbd_settings.get("security").get("cluster_public_key_file"))


def test_node_connect_ok_to_secure_cluster_with_file(qdbd_settings):
    quasardb.Node(
        uri=qdbd_settings.get("uri").get("secure").replace("qdb://", ""),
        user_security_file=qdbd_settings.get("security").get("user_private_key_file"),
        cluster_public_key_file=qdbd_settings.get("security").get("cluster_public_key_file"))


def test_node_empty_prefix(qdbd_direct_connection):
    res = qdbd_direct_connection.prefix_get("testazeazeaze", 10)
    assert len(res) == 0


def test_node_find_prefixes(qdbd_direct_connection):
    res = qdbd_direct_connection.prefix_get("$qdb.statistics.", 20)
    assert len(res) > 0


def test_node_direct_integer(qdbd_direct_connection, entry_name, random_integer):
    entry = qdbd_direct_connection.integer(entry_name)

    with pytest.raises(quasardb.AliasNotFoundError):
        entry.get()

    entry.put(random_integer)
    assert entry.get() == random_integer

    entry.update(random_integer + 1)
    assert entry.get() == random_integer + 1

    entry.remove()

    with pytest.raises(quasardb.AliasNotFoundError):
        entry.get()


def test_node_direct_blob(qdbd_direct_connection, entry_name, random_blob):
    entry = qdbd_direct_connection.blob(entry_name)

    with pytest.raises(quasardb.AliasNotFoundError):
        entry.get()

    entry.put(random_blob)
    assert entry.get() == random_blob

    entry.update(random_blob + random_blob)
    assert entry.get() == random_blob + random_blob

    entry.remove()

    with pytest.raises(quasardb.AliasNotFoundError):
        entry.get()



# pylint: disable=C0103,C0111,C0302,W0212
import datetime
import pytest
import numpy as np
import quasardb


def _generate_dates(start_time, count):
    return np.array([(start_time + np.timedelta64(i, 's'))
                     for i in range(count)]).astype('datetime64[ns]')


def _generate_double_ts(start_time, count):
    return (_generate_dates(start_time, count),
            np.random.uniform(-100.0, 100.0, count))


def _generate_int64_ts(start_time, count):
    return (_generate_dates(start_time, count),
            np.random.randint(-100, 100, count))

def _generate_timestamp_ts(start_time, start_val, count):
    return (
        _generate_dates(start_time, count),
        _generate_dates(start_val, count))

def _generate_blob_ts(start_time, count):
    dates = _generate_dates(start_time, count)
    values = np.array(
        list(
            np.random.bytes(
                np.random.randint(
                    16,
                    32)) for i in range(count)),
        dtype=np.object_)
    return (dates, values)


def _generate_string_ts(start_time, count):
    dates = _generate_dates(start_time, count)
    values = np.array([("content_" + str(item)) for item in range(count)],
                      dtype=np.dtype('U'))

    return (dates, values)

def _check_ts_results(results, generated, count):
    assert len(results) == 2

    np.testing.assert_array_equal(results[0][:count], generated[0][:count])
    np.testing.assert_array_equal(results[1][:count], generated[1][:count])

def test_column_info_repr(column_name):
    double = quasardb.ColumnInfo(quasardb.ColumnType.Double, column_name)
    blob = quasardb.ColumnInfo(quasardb.ColumnType.Blob, column_name)
    int64 = quasardb.ColumnInfo(quasardb.ColumnType.Int64, column_name)
    timestamp = quasardb.ColumnInfo(quasardb.ColumnType.Timestamp, column_name)
    string = quasardb.ColumnInfo(quasardb.ColumnType.String, column_name)
    symbol = quasardb.ColumnInfo(quasardb.ColumnType.Symbol, column_name, 'symtable')

    assert column_name in str(double)
    assert column_name in str(blob)
    assert column_name in str(int64)
    assert column_name in str(timestamp)
    assert column_name in str(string)
    assert column_name in str(symbol)

    assert 'double'    in str(double)
    assert 'blob'      in str(blob)
    assert 'int64'     in str(int64)
    assert 'timestamp' in str(timestamp)
    assert 'string'    in str(string)
    assert 'symbol'    in str(symbol)


def test_list_columns_throws_when_timeseries_does_not_exist(
        qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    with pytest.raises(quasardb.AliasNotFoundError):
        table.list_columns()


def test_insert_throws_when_timeseries_does_not_exist(
        qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)

    with pytest.raises(quasardb.AliasNotFoundError):
        table.double_insert("the_double",
                            np.array(np.datetime64('2011-01-01', 'ns')),
                            np.array([1.0]))


def test_get_ranges_throws_when_timeseries_does_not_exist(
        qdbd_connection, entry_name, intervals):
    table = qdbd_connection.table(entry_name)
    with pytest.raises(quasardb.AliasNotFoundError):
        table.double_get_ranges("blah", intervals)


def test_erase_ranges_throw_when_timeseries_does_not_exist(
        qdbd_connection, entry_name, intervals):
    table = qdbd_connection.table(entry_name)
    with pytest.raises(quasardb.AliasNotFoundError):
        table.erase_ranges("blah", intervals)


def test_create_without_columns(qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    table.create([])

    assert len(table.list_columns()) == 0


def test_create_with_shard_size_less_than_1_millisecond_throws(
        qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    with pytest.raises(quasardb.InvalidArgumentError):
        table.create([], datetime.timedelta(milliseconds=0))


def test_create_with_shard_size_of_1_millisecond(qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    table.create([], datetime.timedelta(milliseconds=1))
    assert len(table.list_columns()) == 0


def test_create_with_shard_size_of_1_day(qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    table.create([], datetime.timedelta(hours=24))
    assert len(table.list_columns()) == 0


def test_create_with_shard_size_of_4_weeks(qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    table.create([], datetime.timedelta(weeks=4))
    assert len(table.list_columns()) == 0


def test_create_with_shard_size_of_more_than_1_year(
        qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    table.create([], datetime.timedelta(weeks=52))
    assert len(table.list_columns()) == 0


def test_table_layout(table):
    """
    We have hard-coded a certain table layout, this test ensures that this is true.
    """
    expected = [("the_double", quasardb.ColumnType.Double),
                ("the_blob", quasardb.ColumnType.Blob),
                ("the_string", quasardb.ColumnType.String),
                ("the_int64", quasardb.ColumnType.Int64),
                ("the_ts", quasardb.ColumnType.Timestamp),
                ("the_symbol", quasardb.ColumnType.Symbol)]

    col_list = table.list_columns()
    got = [(c.name, c.type) for c in col_list]

    assert expected == got

def test_column_lookup(table):
    assert table.column_index_by_id("the_double") == 0
    assert table.column_index_by_id("the_blob") == 1
    assert table.column_index_by_id("the_string") == 2
    assert table.column_index_by_id("the_int64") == 3
    assert table.column_index_by_id("the_ts") == 4

    with pytest.raises(quasardb.Error):
        table.column_index_by_id('foobar')

    with pytest.raises(TypeError):
        table.column_index_by_id(None)

    with pytest.raises(TypeError):
        table.column_index_by_id()


def test_cannot_double_create(table, entry_name):
    with pytest.raises(quasardb.AliasAlreadyExistsError):
        table.create([quasardb.ColumnInfo(
            quasardb.ColumnType.Double, entry_name)])


def _double_col_name(table):
    return table.list_columns()[0].name


def _blob_col_name(table):
    return table.list_columns()[1].name


def _string_col_name(table):
    return table.list_columns()[2].name


def _int64_col_name(table):
    return table.list_columns()[3].name


def _ts_col_name(table):
    return table.list_columns()[4].name

def _symbol_col_name(table):
    return table.list_columns()[5].name

def _start_time(intervals):
    return intervals[0][0]


def _start_year(intervals):
    return intervals[0][0]



# pylint: disable=C0103,C0111,C0302,W0212

import conftest
import quasardb
import pytest

import test_batch_inserter as batchlib


# Don't use `qdbd_connection` fixture, as it's module-scoped (for perf reasons)
# and our options modifications would persist across tests.
#
# XXX(leon): is there an easy way to tell pytest, "please use function-scope
#            for these specific tests" ?
def test_set_client_max_in_buf_1(qdbd_settings):
    with conftest.create_qdbd_connection(qdbd_settings) as conn:
        with pytest.raises(quasardb.InvalidArgumentError):
            conn.options().set_client_max_in_buf_size(1)


# Don't use `qdbd_connection` fixture, as it's module-scoped (for perf reasons)
# and our options modifications would persist across tests.
#
# XXX(leon): is there an easy way to tell pytest, "please use function-scope
#            for these specific tests" ?
def test_set_client_max_in_buf_1MB(qdbd_settings):
    with conftest.create_qdbd_connection(qdbd_settings) as conn:
        conn.options().set_client_max_in_buf_size(1024 * 1024)
        assert conn.options().get_client_max_in_buf_size() == (1024 * 1024)


def test_get_client_max_in_buf(qdbd_connection):
    assert qdbd_connection.options().get_client_max_in_buf_size() > 0


def test_get_cluster_max_in_buf(qdbd_connection):
    assert qdbd_connection.options().get_cluster_max_in_buf_size() > 0


# Don't use `qdbd_connection` fixture, as it's module-scoped (for perf reasons)
# and our options modifications would persist across tests.
#
# XXX(leon): is there an easy way to tell pytest, "please use function-scope
#            for these specific tests" ?
def test_client_query_buf_size_error(qdbd_settings, table, many_intervals):
    with conftest.create_qdbd_connection(qdbd_settings) as conn:
        # First insert some data
        inserter = conn.inserter(
            batchlib._make_inserter_info(table))

        # doubles, blobs, strings, integers, timestamps =
        # batchlib._test_with_table(
        batchlib._test_with_table(
            inserter,
            table,
            many_intervals,
            batchlib._regular_push)

        res = conn.query("select * from \"" + table.get_name() + "\"")

        assert len(res) == 10000

        conn.options().set_client_max_in_buf_size(1500)

        with pytest.raises(quasardb.InputBufferTooSmallError, match=r'consider increasing the buffer size'):
            conn.query("select * from \"" + table.get_name() + "\"")



# # pylint: disable=C0103,C0111,C0302,W0212
import pytest
import quasardb
import numpy as np

import test_table as tslib


def _insert_double_points(table, start_time, points=10):
    inserted_double_data = tslib._generate_double_ts(start_time, points)
    table.double_insert(tslib._double_col_name(table),
                        inserted_double_data[0],
                        inserted_double_data[1])
    return inserted_double_data


def _insert_blob_points(table, start_time, points=10):
    inserted_blob_data = tslib._generate_blob_ts(start_time, points)
    table.blob_insert(tslib._blob_col_name(table),
                      inserted_blob_data[0],
                      inserted_blob_data[1])
    return inserted_blob_data

def _insert_string_points(table, start_time, points=10):
    xs = tslib._generate_string_ts(start_time, points)
    table.string_insert(tslib._string_col_name(table),
                        xs[0],
                        xs[1])
    return xs


def _insert_int64_points(table, start_time, points=10):
    inserted_int64_data = tslib._generate_int64_ts(start_time, points)
    table.int64_insert(tslib._int64_col_name(table),
                       inserted_int64_data[0],
                       inserted_int64_data[1])
    return inserted_int64_data


def _insert_timestamp_points(table, start_time, points=10):
    inserted_timestamp_data = tslib._generate_timestamp_ts(
        start_time, start_time, points)
    table.timestamp_insert(tslib._ts_col_name(table),
                           inserted_timestamp_data[0],
                           inserted_timestamp_data[1])
    return inserted_timestamp_data


def _insert_symbol_points(table, start_time, points=10):
    inserted_symbol_data = tslib._generate_symbol_ts(
        start_time, points)
    table.symbol_insert(tslib._ts_col_name(table),
                        inserted_symbol_data[0],
                        inserted_symbol_data[1])
    return inserted_symbol_data

point_inserter_by_type = {'double': _insert_double_points,
                          'blob': _insert_blob_points,
                          'string': _insert_string_points,
                          'int64': _insert_int64_points,
                          'timestamp': _insert_timestamp_points,
                          'symbol': _insert_symbol_points}

def _insert_points(value_type, table, start_time=None, intervals=None, points=10):
    if start_time is None:
        assert intervals is not None
        start_time = tslib._start_time(intervals)

    assert start_time is not None

    fn = point_inserter_by_type[value_type]
    return fn(table, start_time, points)


def _column_name(table, value_type):
    value_type_to_fn = {'double': tslib._double_col_name,
                        'int64': tslib._int64_col_name,
                        'blob': tslib._blob_col_name,
                        'string': tslib._string_col_name,
                        'timestamp': tslib._ts_col_name,
                        'symbol': tslib._symbol_col_name}

    fn = value_type_to_fn[value_type]
    return fn(table)


##
# Query failure tests


def test_returns_invalid_argument_for_null_query(qdbd_connection):
    with pytest.raises(TypeError):
        qdbd_connection.query()


def test_returns_invalid_argument_for_empty_query(qdbd_connection):
    with pytest.raises(quasardb.InvalidQueryError):
        qdbd_connection.query('')


def test_returns_invalid_argument_with_invalid_query(qdbd_connection):
    with pytest.raises(quasardb.InvalidQueryError):
        qdbd_connection.query('select * from')


def test_returns_alias_not_found_when_ts_doesnt_exist(qdbd_connection):
    with pytest.raises(quasardb.AliasNotFoundError):
        qdbd_connection.query(
            'select * from ' +
            'this_ts_doesnt_exist' +
            ' in range(2017, +10d)')


def test_returns_alias_not_found_when_untagged(qdbd_connection, tag_name):
    with pytest.raises(quasardb.AliasNotFoundError):
        qdbd_connection.query(
            "select * from find(tag='" + tag_name + "') in range(2017, +10d)")


def test_returns_columns_not_found(qdbd_connection, table, column_name):
    with pytest.raises(quasardb.Error):
        qdbd_connection.query(
            "select " +
            column_name +
            " from \"" +
            table.get_name() +
            "\" in range(2017, +10d)")

def test_invalid_query_message(qdbd_connection, table):
    with pytest.raises(quasardb.Error) as excinfo:
        qdbd_connection.query(
            "select asd_col from \"" +
            table.get_name() +
            "\" in range(2016-01-01 , 2016-12-12)")
    assert str(excinfo.value) == 'at qdb_query: The timeseries does not contain this column.'

##
# Double data tests

def sanity_check(ts_name, scanned_point_count, res, rows_count, columns_count):
    assert res.scanned_point_count == scanned_point_count
    assert len(res.tables) == 1
    assert len(res.tables[ts_name][0].data) == rows_count
    assert len(res.tables[ts_name]) == columns_count
    assert res.tables[ts_name][0].name == "$timestamp"


def test_returns_empty_result(qdbd_connection, table):
    res = qdbd_connection.query(
        "select * from \"" +
        table.get_name() +
        "\" in range(2016-01-01 , 2016-12-12)")
    assert len(res) == 0


def test_returns_table_as_string(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select * from \"" + table.get_name() + "\" in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query)

    assert len(res) == 10

    for row, _ in zip(res, inserted_double_data[1]):
        assert row['$table'] == table.get_name()


def test_returns_table_as_blob(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select * from \"" + table.get_name() + "\" in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query, blobs=['$table'])

    assert len(res) == 10

    for row, _ in zip(res, inserted_double_data[1]):
        assert row['$table'] == table.get_name()


def test_returns_inserted_data_with_star_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select * from \"" + table.get_name() + "\" in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query)

    assert len(res) == 10

    for row, v in zip(res, inserted_double_data[1]):
        assert '$timestamp' in row

        assert row['the_blob'] is None
        assert row['the_int64'] is None
        assert row['the_ts'] is None
        assert row['$table'] == table.get_name()
        assert row['the_double'] == v



@pytest.mark.parametrize("query_handler", ['dict',
                                           'numpy'])
@pytest.mark.parametrize("value_type", ['double',
                                        'int64',
                                        'blob',
                                        'string',
                                        'timestamp'])
def test_supports_all_column_types(value_type, query_handler, qdbd_connection, table, intervals):
    inserted_data = _insert_points(value_type, table, intervals=intervals)
    column_name = _column_name(table, value_type)
    query = "SELECT \"{}\" FROM \"{}\"".format(column_name, table.get_name())

    if query_handler == 'numpy':
        res = qdbd_connection.query_numpy(query)
        (col, xs) = res[0]
        assert col == column_name
        np.testing.assert_array_equal(xs,
                                      inserted_data[1])
    elif query_handler == 'dict':
        res = qdbd_connection.query(query)
        assert len(res) == len(inserted_data[1])

        for i in range(len(res)):
            assert column_name in res[i]

            val_in  = inserted_data[1][i]
            val_out = res[i][column_name]

            assert val_in == val_out

    else:
        raise RuntimeError("Unrecognized query handler: {}".format(query_handler))


@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
@pytest.mark.parametrize("query_handler", ['dict',
                                           'numpy'])
@pytest.mark.parametrize("value_type", ['double',
                                        'int64',
                                        'blob',
                                        'string',
                                        'timestamp'])
@pytest.mark.parametrize("row_count", [16384, 131072, 1048576])
def test_query_handler_benchmark(benchmark, value_type, row_count, query_handler, qdbd_connection, table, intervals):



    query_fn = None
    if query_handler == 'numpy':
        query_fn = qdbd_connection.query_numpy
    elif query_handler == 'dict':
        query_fn = qdbd_connection.query
    else:
        raise RuntimeError("Unrecognized query handler: {}".format(query_handler))



    inserted_data = _insert_points(value_type, table, intervals=intervals, points=row_count)
    column_name = _column_name(table, value_type)
    query = "SELECT \"{}\" FROM \"{}\"".format(column_name, table.get_name())

    benchmark(query_fn, query)



def test_returns_inserted_data_with_column_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select " + tslib._double_col_name(table) + \
        " from \"" + table.get_name() + "\" in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"

    res = qdbd_connection.query(query)

    assert len(res) == 10

    for row, v in zip(res, inserted_double_data[1]):
        assert '$table' not in row
        assert '$timestamp' not in row
        assert 'the_blob' not in row
        assert 'the_int64' not in row
        assert 'the_ts' not in row

        assert row['the_double'] == v

def test_returns_inserted_data_with_specific_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select $timestamp, $table, " + tslib._double_col_name(table) + \
        " from \"" + table.get_name() + "\" in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query)

    assert len(res) == 10

    for row, v in zip(res, inserted_double_data[1]):
        assert '$timestamp' in row
        assert 'the_blob' not in row
        assert 'the_int64' not in row
        assert 'the_ts' not in row

        assert row['the_double'] == v


def test_returns_count_data_with_count_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    _ = _insert_double_points(table, start_time, 10)
    query = "select count(" + tslib._double_col_name(table) + ") from \"" + table.get_name() + \
        "\" in range(" + str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query)

    assert len(res) == 1
    assert res[0]['count(the_double)'] == 10


def test_returns_count_data_with_sum_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select sum(" + tslib._double_col_name(table) + ") from \"" + table.get_name() + \
        "\" in range(" + str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query)

    assert len(res) == 1
    assert res[0]['sum(the_double)'] == pytest.approx(np.sum(inserted_double_data[1]))


def test_returns_inserted_multi_data_with_star_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 100)
    inserted_blob_data = _insert_blob_points(table, start_time, 100)
    inserted_int64_data = _insert_int64_points(table, start_time, 100)
    inserted_timestamp_data = _insert_timestamp_points(table, start_time, 100)

    query = "select * from \"" + table.get_name() + \
        "\" in range(" + str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query, blobs=['the_blob'])

    assert len(res) == 100

    for row, double, blob, int64, ts in zip(res,
                                            inserted_double_data[1],
                                            inserted_blob_data[1],
                                            inserted_int64_data[1],
                                            inserted_timestamp_data[1]):
        assert '$timestamp' in row

        # Note that this is a string
        assert row['$table'] == table.get_name()

        # And this is a blob
        assert row['the_blob'] == blob

        assert row['the_double'] == double
        assert row['the_int64'] == int64
        assert row['the_ts'] == ts


def test_create_table(qdbd_connection, entry_name):
    query = "create table \"{}\" (col int64)".format(entry_name)
    _ = qdbd_connection.query(query)



# pylint: disable=missing-module-docstring,missing-function-docstring,not-an-iterable,invalid-name
import pytest
import quasardb
import numpy as np
import test_table as tslib
import time
import datetime

def _insert_double_points(table, start_time, points=10):
    inserted_double_data = tslib._generate_double_ts(start_time, points)
    table.double_insert(tslib._double_col_name(table),
                        inserted_double_data[0],
                        inserted_double_data[1])
    return inserted_double_data

def test_returns_invalid_argument_for_null_query_full(qdbd_connection):
    with pytest.raises(TypeError):
        qdbd_connection.query_continuous_full()

def test_returns_invalid_argument_for_null_query_full(qdbd_connection):
    with pytest.raises(TypeError):
        qdbd_connection.query_continuous_new_values()

def test_returns_empty_result_full(qdbd_connection, table):
    # return empty result
    q = "select * from \"" + table.get_name() + "\""
    cont = qdbd_connection.query_continuous_full(q, datetime.timedelta(milliseconds=100))
    res = cont.results()
    assert len(res) == 0

def test_returns_empty_result_new_values(qdbd_connection, table):
    # return empty result
    q = "select * from \"" + table.get_name() + "\""
    cont = qdbd_connection.query_continuous_new_values(q, datetime.timedelta(milliseconds=100))
    res = cont.results()
    assert len(res) == 0

def test_returns_empty_result_new_values_probe(qdbd_connection, table):
    # return empty result
    q = "select * from \"" + table.get_name() + "\""
    cont = qdbd_connection.query_continuous_new_values(q, datetime.timedelta(milliseconds=100))
    res = cont.probe_results()
    assert len(res) == 0

def _test_against_table(res, table, data):
    for row, v in zip(res, data):
        assert '$timestamp' in row

        assert row['the_blob'] == None
        assert row['the_int64'] == None
        assert row['the_ts'] == None
        assert row['$table'] == table.get_name()
        assert row['the_double'] == v

def test_returns_rows_full(qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 1)
    q = "select * from \"" + table.get_name() + "\""
    cont = qdbd_connection.query_continuous_full(q, datetime.timedelta(milliseconds=100))
    res = cont.results()
    assert len(res) == 1
    _test_against_table(res, table, inserted_double_data[1])

def test_returns_rows_full_probe(qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 1)
    q = "select * from \"" + table.get_name() + "\""
    cont = qdbd_connection.query_continuous_full(q, datetime.timedelta(milliseconds=100))
    res = []
    while len(res) == 0:
        res = cont.probe_results()
    assert len(res) == 1
    _test_against_table(res, table, inserted_double_data[1])

def test_returns_rows_full_iterator(qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 1)
    q = "select * from \"" + table.get_name() + "\""
    cont = qdbd_connection.query_continuous_full(q, datetime.timedelta(milliseconds=100))
    for res in cont:
        assert len(res) == 1
        _test_against_table(res, table, inserted_double_data[1])
        break

def test_returns_rows_new_values(qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 1)
    q = "select * from \"" + table.get_name() + "\""
    cont = qdbd_connection.query_continuous_new_values(q, datetime.timedelta(milliseconds=100))
    res = cont.results()
    assert len(res) == 1
    _test_against_table(res, table, inserted_double_data[1])

def test_returns_rows_new_values_probe(qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 1)
    q = "select * from \"" + table.get_name() + "\""
    cont = qdbd_connection.query_continuous_new_values(q, datetime.timedelta(milliseconds=100))
    res = []
    while len(res) == 0:
        res = cont.probe_results()
    assert len(res) == 1
    _test_against_table(res, table, inserted_double_data[1])

def test_returns_rows_new_value_iterator(qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 1)
    q = "select * from \"" + table.get_name() + "\""
    cont = qdbd_connection.query_continuous_new_values(q, datetime.timedelta(milliseconds=100))
    for res in cont:
        assert len(res) == 1
        _test_against_table(res, table, inserted_double_data[1])
        break


def __wait_for(cont, f, max_ticks=10):
    i = 0

    for res in cont:

        v = f(res)
        if v:
            return v

        i += 1

        # we must not wait for the value too long
        assert i < max_ticks


def test_returns_rows_full_value_iterator_multiple(qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 1)
    q = "select * from \"" + table.get_name() + "\""
    cont = qdbd_connection.query_continuous_full(q, datetime.timedelta(milliseconds=100))

    assert True is __wait_for(cont, lambda x: len(x) == 1)

    start_time += np.timedelta64(1, 'm')
    inserted_double_data = _insert_double_points(table, start_time, 1)

    assert True is __wait_for(cont, lambda x: len(x) == 2)


def test_returns_rows_new_value_iterator_multiple(qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 1)
    q = "select * from \"" + table.get_name() + "\""
    cont = qdbd_connection.query_continuous_new_values(q, datetime.timedelta(milliseconds=100))

    assert True is __wait_for(cont, lambda x: len(x) == 1)

    start_time += np.timedelta64(1, 'm')
    inserted_double_data = _insert_double_points(table, start_time, 1)

    assert True is __wait_for(cont, lambda x: len(x) == 1)



import logging

logger = logging.getLogger('quasardb.table_cache')

_cache = {}

def clear():
    logger.info("Clearing table cache")
    _cache = {}

def exists(table_name: str) -> bool:
    """
    Returns true if table already exists in table cache.
    """
    return table_name in _cache

def store(table, table_name=None, force_retrieve_metadata=True):
    """
    Stores a table into the cache. Ensures metadata is retrieved. This is useful if you want
    to retrieve all table metadata at the beginning of a process, to avoid doing expensive
    lookups in undesired code paths.

    Returns a reference to the table being stored.
    """
    if table_name is None:
        table_name = table.get_name()

    if exists(table_name):
        logger.warn("Table already in cache, overwriting: %s", table_name)

    logger.debug("Caching table %s", table_name)
    _cache[table_name] = table

    table.retrieve_metadata()

    return table

def lookup(table_name: str, conn, force_retrieve_metadata=True):
    """
    Retrieves table from _cache if already exists. If it does not exist,
    looks up the table from `conn` and puts it in the cache.

    If `force_retrieve_metadata` equals True, we will ensure that the table's
    metadata is
    """
    if exists(table_name):
        return _cache[table_name]

    logger.debug("table %s not yet found, looking up", table_name)
    table = conn.table(table_name)

    return store(table, table_name, force_retrieve_metadata=force_retrieve_metadata)



# pylint: disable=C0103,C0111,C0302,W0212
import datetime
import pytest
import quasardb


def test_expires_at(blob_entry, random_blob, datetime_, timedelta):
    """
    Test for expiry.
    We want to make sure, in particular, that the conversion from Python datetime is right.
    """

    # entry does not exist yet
    with pytest.raises(quasardb.AliasNotFoundError):
        blob_entry.get_expiry_time()

    blob_entry.put(random_blob)

    exp = blob_entry.get_expiry_time()
    assert isinstance(exp, datetime.datetime)
    assert exp.year == 1970

    future_exp = datetime_ + timedelta
    blob_entry.expires_at(future_exp)

    output_exp = blob_entry.get_expiry_time()
    assert isinstance(output_exp, datetime.datetime)
    assert isinstance(output_exp.tzinfo, datetime.timezone)

    check_exp = future_exp
    if check_exp.tzinfo is None:
        # Per conftest.datetime(), the only "bare" datetime we create is one
        # of `datetime.datetime.now()`, which is local time. As such, as
        # QuasarDB always emits UTC datetimes, we should cast our input time
        # to UTC as well before comparing.
        check_exp = check_exp.astimezone(datetime.timezone.utc)

    assert check_exp == output_exp


def test_expires_from_now(blob_entry, random_blob, timedelta):
    # entry does not exist yet
    with pytest.raises(quasardb.AliasNotFoundError):
        blob_entry.get_expiry_time()

    blob_entry.put(random_blob)

    exp = blob_entry.get_expiry_time()
    assert isinstance(exp, datetime.datetime)
    assert exp.year == 1970

    # expires in one minute from now
    blob_entry.expires_from_now(timedelta)

    # We use a wide 10s interval for the check, because we have no idea at which speed
    # these testcases may run in debug. This will be enough however to check that
    # the interval has properly been converted and the time zone is
    # correct.
    future_exp_lower_bound = datetime.datetime.now(tz=datetime.timezone.utc) + timedelta - datetime.timedelta(seconds=30)
    future_exp_higher_bound = future_exp_lower_bound + timedelta + datetime.timedelta(seconds=30)

    exp = blob_entry.get_expiry_time()
    assert isinstance(exp, datetime.datetime)
    assert future_exp_lower_bound < exp
    assert future_exp_higher_bound > exp



# pylint: disable=C0103,C0111,C0302,W0212
import quasardb


def test_get_name(qdbd_connection, entry_name):
    blob = qdbd_connection.blob(entry_name)
    assert entry_name == blob.get_name()

    integer = qdbd_connection.integer(entry_name)
    assert entry_name == integer.get_name()



# pylint: disable=C0103,C0111,C0302,W0212
import quasardb


def test_node_status(qdbd_connection, qdbd_settings):
    status = qdbd_connection.node_status(
        qdbd_settings.get('uri').get('insecure'))
    assert len(status) > 0
    assert status.get('overall') is not None


def test_node_config(qdbd_connection, qdbd_settings):
    config = qdbd_connection.node_config(
        qdbd_settings.get('uri').get('insecure'))
    assert len(config) > 0
    assert config.get('global') is not None
    assert config.get('local') is not None


def test_node_topology(qdbd_connection, qdbd_settings):
    topology = qdbd_connection.node_topology(
        qdbd_settings.get('uri').get('insecure'))

    assert len(topology) > 0

    assert topology.get('predecessor') is not None
    assert topology.get('center') is not None
    assert topology.get('successor') is not None


def test_endpoints(qdbd_connection, qdbd_settings):
    endpoints = qdbd_connection.endpoints()
    node_endpoint = qdbd_settings.get('uri').get(
        'insecure').replace("qdb://", "")

    assert endpoints == [node_endpoint]



# pylint: disable=missing-module-docstring,missing-function-docstring,not-an-iterable,invalid-name

from datetime import datetime, timedelta
import logging
import numpy as np
import numpy.ma as ma
import pandas as pd
import pytest
import conftest

import quasardb
import quasardb.pandas as qdbpd

ROW_COUNT = 1000

logger = logging.getLogger("test-pandas")

def _to_numpy_masked(xs):
    data = xs.to_numpy()
    mask = xs.isna()
    return ma.masked_array(data=data, mask=mask)

def _assert_series_equal(lhs, rhs):
    lhs_ = _to_numpy_masked(lhs)
    rhs_ = _to_numpy_masked(rhs)

    assert ma.count_masked(lhs_) == ma.count_masked(rhs_)

    logger.debug("lhs: %s", lhs_[:10])
    logger.debug("rhs: %s", rhs_[:10])

    lhs_ = lhs_.torecords()
    rhs_ = rhs_.torecords()

    for ((lval, lmask), (rval, rmask)) in zip(lhs_, rhs_):
        assert lmask == rmask

        if lmask is False:
            assert lval == rval


def _assert_df_equal(lhs, rhs):
    """
    Verifies DataFrames lhs and rhs are equal(ish). We're not pedantic that we're comparing
    metadata and things like that.

    Typically one would use `lhs` for the DataFrame that was generated in code, and
    `rhs` for the DataFrame that's returned by qdbpd.
    """

    np.testing.assert_array_equal(lhs.index.to_numpy(), rhs.index.to_numpy())
    assert len(lhs.columns) == len(rhs.columns)
    for col in lhs.columns:
        _assert_series_equal(lhs[col], rhs[col])


def gen_idx(start_time, count, step=1, unit='D'):
    return pd.Index(data=pd.date_range(start=start_time,
                                       end=(start_time + (np.timedelta64(step, unit) * count)),
                                       periods=count),
                    dtype='datetime64[ns]',
                    name='$timestamp')


def gen_df(start_time, count, step=1, unit='D'):
    idx = gen_idx(start_time, count, step, unit)

    return pd.DataFrame(data={"the_double": np.random.uniform(-100.0, 100.0, count),
                              "the_int64": np.random.randint(-100, 100, count),
                              "the_blob": np.array(list(np.random.bytes(np.random.randint(16, 32)) for i in range(count)),
                                                   'O'),
                              "the_string": np.array([("content_" + str(item))
                                                      for item in range(count)], 'U'),
                              "the_ts": np.array([
                                  (start_time + np.timedelta64(i, 's'))
                                  for i in range(count)]).astype('datetime64[ns]'),
                              "the_symbol": np.array([("symbol_" + str(item))
                                                      for item in range(count)], 'U')},

                        index=idx)


def gen_series(start_time, count):
    idx = pd.date_range(start_time, periods=count, freq='S')

    return {"the_double": pd.Series(np.random.uniform(-100.0, 100.0, count),
                                    index=idx),
            "the_int64": pd.Series(np.random.randint(-100, 100, count),
                                   index=idx),
            "the_blob": pd.Series(np.array(list(np.random.bytes(np.random.randint(16, 32)) for i in range(count)),
                                           'O'),
                                  index=idx),
            "the_string": pd.Series(np.array([("content_" + str(item)) for item in range(count)],
                                             'U'),
                                    index=idx),
            "the_ts": pd.Series(np.array([(start_time + np.timedelta64(i, 's'))
                                          for i in range(count)]).astype('datetime64[ns]'),
                                index=idx),
            "the_symbol": pd.Series(np.array([("symbol_" + str(item)) for item in range(count)],
                                             'U'),
                                    index=idx)}


def test_series_read_write(series_with_table):
    (ctype, dtype, series, table) = series_with_table

    col = table.column_id_by_index(0)
    qdbpd.write_series(series, table, col, dtype=dtype)

    # Read everything
    res = qdbpd.read_series(table, col)

    _assert_series_equal(series, res)


def test_dataframe(qdbpd_write_fn, df_with_table, qdbd_connection):
    (_, _, df1, table) = df_with_table
    qdbpd_write_fn(df1, qdbd_connection, table)

    df2 = qdbpd.read_dataframe(table)

    _assert_df_equal(df1, df2)

@pytest.mark.parametrize('sparsify', conftest.no_sparsify)
def test_dataframe_can_read_columns(qdbpd_write_fn, df_with_table, qdbd_connection, column_name, table_name):
    (ctype, dtype, df1, table) = df_with_table
    assert column_name in df1.columns
    assert len(df1.columns) == 1

    qdbpd_write_fn(df1, qdbd_connection, table, infer_types=False, dtype=dtype)

    df2 = qdbpd.read_dataframe(table, columns=[column_name])

    _assert_df_equal(df1, df2)


def test_dataframe_can_read_ranges(qdbpd_write_fn, qdbd_connection, df_with_table, start_date, row_count):
    (ctype, dtype, df1, table) = df_with_table

    qdbpd_write_fn(df1, qdbd_connection, table)

    offset_25 = start_date + np.timedelta64(int(row_count / 4), 's')
    offset_75 = start_date + (3 * np.timedelta64(int(row_count / 4), 's'))

    range_0_25 = (start_date, offset_25)
    range_25_75 = (offset_25, offset_75)

    df2 = qdbpd.read_dataframe(table)
    df3 = qdbpd.read_dataframe(table, ranges=[range_0_25])
    df4 = qdbpd.read_dataframe(table, ranges=[range_25_75])
    df5 = qdbpd.read_dataframe(table, ranges=[range_0_25, range_25_75])

    assert df2.shape[0] == row_count
    assert df3.shape[0] == row_count / 4
    assert df4.shape[0] == row_count / 2
    assert df5.shape[0] == (row_count / 4) * 3

def test_write_dataframe(qdbpd_write_fn, df_with_table, qdbd_connection):
    (ctype, dtype, df, table) = df_with_table

    # We always need to infer
    qdbpd_write_fn(df, qdbd_connection, table, infer_types=True)
    res = qdbpd.read_dataframe(table)

    _assert_df_equal(df, res)


def test_multiple_dataframe(qdbpd_writes_fn, dfs_with_tables, qdbd_connection):

    dfs = []

    for (_, _, df, table) in dfs_with_tables:
        dfs.append((table.get_name(), df))

    qdbpd_writes_fn(dfs, qdbd_connection, infer_types=True, fast=True)

    for (_, _, df, table) in dfs_with_tables:
        res = qdbpd.read_dataframe(table)
        _assert_df_equal(df, res)

# Pandas is a retard when it comes to null values, so make sure to just only
# generate "full" arrays when we don't infer / convert array types.
@pytest.mark.parametrize('sparsify', conftest.no_sparsify)
def test_write_dataframe_no_infer(qdbpd_write_fn, df_with_table, qdbd_connection):
    (ctype, dtype, df, table) = df_with_table

    # We always need to infer
    if dtype == np.str_:
        logger.warn("Skipping test because pandas messes up strings")
        with pytest.raises(TypeError, match=r'Data for column \'([a-z]{16})\' with type \'ColumnType.String\' was provided in dtype \''):
            qdbpd_write_fn(df, qdbd_connection, table, infer_types=False)
    else:
        qdbpd_write_fn(df, qdbd_connection, table, infer_types=False)
        res = qdbpd.read_dataframe(table)

        _assert_df_equal(df, res)


def test_write_unindexed_dataframe(qdbpd_write_fn, df_with_table, df_count,
                                   qdbd_connection, caplog):
    # CF. QDB-10203, we generate a dataframe which is unordered. The easiest
    # way to do this is reuse our gen_df function with overlapping timestamps,
    # and concatenate the data.

    # We generate a whole bunch of dataframes. Their timestamps overlap, but
    # never collide: we ensure that by using a 'step' of the dataframe count,
    # and let each dataframe start at the appropriate offset.
    #
    # e.g. with 8 dataframes, the first dataframe starts at 2017-01-01 and the
    # second row's timestap is 2017-01-09, while the second dataframe starts at
    # 2017-01-02 and has 2017-01-10 as second row, etc.

    (ctype, dtype, df, table) = df_with_table

    dfs = (df.copy(deep=True) for i in range(df_count))
    dfs = (df.sample(frac=1) for df in dfs)

    df_unsorted = pd.concat(dfs)
    df_sorted = df_unsorted.sort_index().reindex()
    assert len(df_sorted.columns) == len(df_unsorted.columns)

    # We store unsorted
    caplog.clear()
    caplog.set_level(logging.WARN, logger='quasardb.pandas')
    qdbpd_write_fn(df_unsorted, qdbd_connection, table)
    msgs = [x.message for x in caplog.records]
    assert "dataframe index is unsorted, resorting dataframe based on index" in msgs

    # We expect to receive sorted data
    df_read = qdbpd.read_dataframe(table)

    _assert_df_equal(df_sorted, df_read)


def test_write_dataframe_push_fast(qdbpd_write_fn, qdbd_connection, df_with_table):
    (_, _, df1, table) = df_with_table

    # Ensures that we can do a full-circle write and read of a dataframe
    qdbpd_write_fn(df1, qdbd_connection, table, fast=True)

    df2 = qdbpd.read_dataframe(table)

    _assert_df_equal(df1, df2)

def test_write_dataframe_push_truncate(qdbpd_write_fn, qdbd_connection, df_with_table):
    (_, _, df1, table) = df_with_table

    # Ensures that we can do a full-circle write and read of a dataframe
    qdbpd_write_fn(df1, qdbd_connection, table, truncate=True)
    qdbpd_write_fn(df1, qdbd_connection, table, truncate=True)

    df2 = qdbpd.read_dataframe(table)

    _assert_df_equal(df1, df2)

def test_write_dataframe_deduplicate(qdbpd_write_fn, qdbd_connection, df_with_table, deduplicate):
    (_, _, df1, table) = df_with_table

    # Main validation:
    # 1. write once, no deduplication
    # 2. write again, with deduplication
    # 3. read data again
    #
    # data from step 3 should be identical to the input dataframe, as all data
    # should be deduplicated upon insertion.

    # Ensures that we can do a full-circle write and read of a dataframe
    qdbpd_write_fn(df1, qdbd_connection, table, deduplicate=False, deduplication_mode='drop')
    qdbpd_write_fn(df1, qdbd_connection, table, deduplicate=deduplicate, deduplication_mode='drop')

    df2 = qdbpd.read_dataframe(table)

    _assert_df_equal(df1, df2)


def test_write_dataframe_create_table(qdbpd_write_fn, qdbd_connection, gen_df, table_name):
    (_, _, df1) = gen_df

    table = qdbd_connection.ts(table_name)
    qdbpd_write_fn(df1, qdbd_connection, table, create=True)

    df2 = qdbpd.read_dataframe(table)

    _assert_df_equal(df1, df2)

def test_write_dataframe_create_table_twice(qdbpd_write_fn, qdbd_connection, df_with_table):
    (_, _, df, table)  = df_with_table
    qdbpd_write_fn(df, qdbd_connection, table, create=True)

def test_write_dataframe_create_table_with_shard_size(qdbpd_write_fn, qdbd_connection, gen_df, table_name):
    (_, _, df1) = gen_df
    table = qdbd_connection.ts(table_name)

    qdbpd_write_fn(df1,
                   qdbd_connection,
                   table,
                   create=True,
                   shard_size=timedelta(weeks=4))

    df2 = qdbpd.read_dataframe(table)

    _assert_df_equal(df1, df2)


def test_query(qdbpd_write_fn, # parametrized
               qdbpd_query_fn, # parametrized
               df_with_table,  # parametrized
               qdbd_connection,
               column_name,
               table_name):
    (_, _, df1, table) = df_with_table
    qdbpd_write_fn(df1, qdbd_connection, table)

    df2 = qdbpd_query_fn(qdbd_connection,
                         "SELECT $timestamp, {} FROM \"{}\"".format(column_name, table_name)
                         ).set_index('$timestamp').reindex()

    _assert_df_equal(df1, df2)


def test_inference(
        qdbpd_write_fn, # parametrized
        df_with_table,
        qdbd_connection):
    # Note that there are no tests; we effectively only test whether it doesn't
    # throw.
    (_, _, df, table) = df_with_table
    qdbpd_write_fn(df, qdbd_connection, table, infer_types=True)


def test_shuffled_columns(qdbpd_write_fn, qdbd_connection, table_name, start_date, row_count):
    t = qdbd_connection.table(table_name)

    # Specific regresion test used in Python user guide / API documentation
    cols = [quasardb.ColumnInfo(quasardb.ColumnType.Double, "open"),
            quasardb.ColumnInfo(quasardb.ColumnType.Double, "close"),
            quasardb.ColumnInfo(quasardb.ColumnType.Double, "high"),
            quasardb.ColumnInfo(quasardb.ColumnType.Double, "low"),
            quasardb.ColumnInfo(quasardb.ColumnType.Int64, "volume")]

    t.create(cols)

    idx = np.array([start_date + np.timedelta64(i, 's')
                    for i in range(row_count)]).astype('datetime64[ns]')

    data = {
        '$timestamp': idx,
        '$table': [table_name for i in range(row_count)],
        'close': np.random.uniform(100, 200, row_count),
        'volume': np.random.randint(10000, 20000, row_count),
        'open': np.random.uniform(100, 200, row_count),
}

    df = pd.DataFrame(data).set_index('$timestamp').reindex()

    qdbpd_write_fn(df, qdbd_connection, t, infer_types=True)


def test_regression_sc11057(qdbd_connection, table_name):
    """
    Our query results (apparently) return datetime64 arrays with a itemsize of 8,
    but a stride size of 16 -- i.e. every other item is "skipped".

    By re-writing query results as a dataframe, we validate that our backend
    properly deals with these arrays.
    """

    idx  = [np.datetime64('2022-09-08T12:00:01', 'ns'),
            np.datetime64('2022-09-08T12:00:02', 'ns'),
            np.datetime64('2022-09-08T12:00:03', 'ns')]
    data = {'record_timestamp': [np.datetime64('2022-09-08T13:00:04', 'ns'),
                                 np.datetime64('2022-09-08T13:00:05', 'ns'),
                                 np.datetime64('2022-09-08T13:00:06', 'ns')],
            'unique_tagname': np.array(['ABC', 'DEF', 'GHI'], dtype='unicode')}
    df = pd.DataFrame(data=data, index=idx)

    qdbpd.write_dataframe(df, qdbd_connection, table_name, create=True, infer_types=True)

    q = 'select $timestamp, record_timestamp, unique_tagname from "{}"'.format(table_name)
    df_ = qdbpd.query(qdbd_connection, q)
    df_ = df_.groupby('unique_tagname').last().set_index('$timestamp')

    qdbpd.write_dataframe(df_, qdbd_connection, '{}_out'.format(table_name), create=True, infer_types=True)

    result = qdbpd.read_dataframe(qdbd_connection.table('{}_out'.format(table_name)))

    _assert_df_equal(df_, result)


def test_regression_sc11084(qdbd_connection, table_name):
    """
    This test validates that, if one array starts with a significant amount of
    NULL values, it works properly. This addresses a bug that our "probe_mask" function
    inside our C++ masked_array implementation considers more than just the first chunk
    of 256 values.
    """

    n_rows = 8192
    start = np.datetime64('2022-09-08T12:00:00', 'ns')
    idx = [start + np.timedelta64(i, 'ns') for i in range(0, n_rows)]

    data1 = np.full(int(n_rows / 2), np.nan, dtype='float64')
    data2 = np.full(int(n_rows / 2), 1.11, dtype='float64')
    data = {'val': np.concatenate([data1, data2])}

    df = pd.DataFrame(data=data, index=idx)
    qdbpd.write_dataframe(df, qdbd_connection, table_name, create=True, infer_types=True)

    df_ = qdbpd.query(qdbd_connection, "select $timestamp, val from \"{}\"".format(table_name), index='$timestamp')

    _assert_df_equal(df, df_)

@pytest.mark.parametrize('sparsify', conftest.no_sparsify)
def test_regression_sc11337(qdbpd_write_fn, df_with_table, qdbd_connection, column_name):
    (ctype, dtype, df1, table) = df_with_table
    assert column_name in df1.columns
    assert len(df1.columns) == 1

    df1 = df1.astype('object')

    df1[column_name][4] = pd.NA

    qdbpd_write_fn(df1, qdbd_connection, table.get_name(), fast=True, infer_types=True)

    df2 = qdbpd.read_dataframe(table, columns=[column_name])

    _assert_df_equal(df1, df2)



# pylint: disable=C0103,C0111,C0302,W0212
import pytest
import numpy as np
import test_batch_inserter as batchlib
import test_pandas as pandaslib
import quasardb.pandas as qdbpd

row_count = 10000


@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
def test_bench_double_series(
        qdbd_connection,
        table,
        many_intervals,
        benchmark):
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    # doubles, blobs, strings, integers, timestamps, symbols =
    # batchlib._test_with_table(
    _, _, _, _, _, _ = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    benchmark(qdbpd.read_series, table, "the_double")


@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
def test_bench_blob_series(qdbd_connection, table, many_intervals, benchmark):
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    # doubles, blobs, strings, integers, timestamps, symbols =
    # batchlib._test_with_table(
    _, _, _, _, _, _ = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    benchmark(qdbpd.read_series, table, "the_blob")


@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
def test_bench_string_series(
        qdbd_connection,
        table,
        many_intervals,
        benchmark):
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    # doubles, blobs, strings, integers, timestamps, symbols =
    # batchlib._test_with_table(
    _, _, _, _, _, _ = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    benchmark(qdbpd.read_series, table, "the_string")


@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
def test_bench_int64_series(qdbd_connection, table, many_intervals, benchmark):
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    # doubles, blobs, strings, integers, timestamps, symbols =
    # batchlib._test_with_table(
    _, _, _, _, _, _ = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    benchmark(qdbpd.read_series, table, "the_int64")


@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
def test_bench_timestamp_series(
        qdbd_connection,
        table,
        many_intervals,
        benchmark):
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    # doubles, blobs, strings, integers, timestamps, symbols =
    # batchlib._test_with_table(
    _, _, _, _, _, _ = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    benchmark(qdbpd.read_series, table, "the_ts")


@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
def test_bench_symbol_series(
        qdbd_connection,
        table,
        many_intervals,
        benchmark):
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    # doubles, blobs, strings, integers, timestamps, symbols =
    # batchlib._test_with_table(
    _, _, _, _, _, _ = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    benchmark(qdbpd.read_series, table, "the_symbol")


@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
def test_benchmark_dataframe_read(qdbd_connection, table, benchmark):
    df = pandaslib.gen_df(np.datetime64('2017-01-01'), row_count)
    qdbpd.write_dataframe(df, qdbd_connection, table)
    benchmark(qdbpd.read_dataframe, table)


@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
def test_benchmark_dataframe_write(qdbd_connection, table, benchmark):
    # Ensures that we can do a full-circle write and read of a dataframe
    df = pandaslib.gen_df(np.datetime64('2017-01-01'), row_count)
    benchmark(qdbpd.write_dataframe, df, qdbd_connection, table)

@pytest.mark.skip(reason="Skip unless you're benching the pinned writer")
@pytest.mark.parametrize("query_handler", ['dict', 'numpy'])
@pytest.mark.parametrize("row_count", [(256),
                                       (256 * 8),
                                       (256 * 8 * 8),
                                       (256 * 8 * 8 * 8),
                                       (256 * 8 * 8 * 8 * 8),])
@pytest.mark.parametrize("column_name", ['the_double',
                                         'the_int64',
                                         'the_blob',
                                         'the_string',
                                         'the_ts'])
def test_benchmark_dataframe_query(query_handler, row_count, column_name, qdbd_connection, table, benchmark):
    qdbd_connection.options().set_client_max_in_buf_size(4 * 1024 * 1024 * 1024)

    # Ensures that we can do a full-circle write and read of a dataframe
    df = pandaslib.gen_df(np.datetime64('2017-01-01'), row_count, unit='s')
    qdbpd.write_pinned_dataframe(df, qdbd_connection, table, fast=True, infer_types=False)

    q = "SELECT \"{}\" FROM \"{}\"".format(column_name,
                                           table.get_name())

    numpy = False
    if query_handler == 'numpy':
        numpy = True

    benchmark(qdbpd.query, qdbd_connection, q, numpy=numpy)



###
#
# XXX: If this import fails, ensure that you built the quasardb python API with
#      export QDB_TESTS_ENABLED=ON. E.g.
#
#      ```
#      export QDB_TESTS_ENABLED=ON
#      python3 setup.py test --addopts "-s tests/test_convert.py"
#      ```

from quasardb import test_convert as m

#
###

from utils import assert_indexed_arrays_equal
import conftest

def test_unicode_u32_decode_traits():
    m.test_unicode_u32_decode_traits()

def test_unicode_u8_encode_traits():
    m.test_unicode_u8_encode_traits()

def test_unicode_u8_decode_traits():
    m.test_unicode_u8_decode_traits()

def test_unicode_u8_recode():
    m.test_unicode_u8_recode()

def test_unicode_decode_algo():
    m.test_unicode_decode_algo()

def _test_array_recode(array_with_index_and_table):
    (ctype, dtype, xs1, idx1, table) = array_with_index_and_table

    if dtype.char == 'S':
        # XXX(leon): we don't yet support native qdb -> np.ndarray with dtype `null-terminated binary` (S). I don't think we
        #            should ever do this, but we do need it for input. That's why we can't test this right now, because we
        #            can't do it full circle.
        return True

    (idx2, xs2) = m.test_array_recode(ctype, dtype, (idx1, xs1))
    assert_indexed_arrays_equal((idx1, xs1), (idx2, xs2))


@conftest.override_sparsify('partial')
def test_array_recode_sparsify_partial(array_with_index_and_table):
    return _test_array_recode(array_with_index_and_table)

@conftest.override_sparsify('none')
def test_array_recode_sparsify_none(array_with_index_and_table):
    return _test_array_recode(array_with_index_and_table)



###
# Verifies properties of our exceptions.
###

import inspect
import pytest
import quasardb
import pprint

classes = inspect.getmembers(quasardb, inspect.isclass)
exception_classes = (v for (k, v) in classes if k.endswith('Error'))
exception_tree = inspect.getclasstree(exception_classes)

def test_base_is_runtime_error():
    assert isinstance(exception_tree, list)

    # The dept should be 4; this test may fail if we increase the nesting
    # level, but I don't anticipate that happening any time soon.
    #
    # 2022-12-12 -- XXX(leon) -- increasing this to 6, because we added an additional
    #            nesting level for assertion errors.
    assert len(exception_tree) == 6

    # Define what we expect the hierarchy to look like
    hierarchy = [Exception, RuntimeError, quasardb.Error]

    cur = exception_tree

    def walk_and_check(xs, depth = 0):
        assert depth <= len(hierarchy)

        # Parent/child class
        if isinstance(xs, tuple):
            # We only expect one base class
            assert len(xs[1]) == 1
            derived_from = xs[1][0]
            if depth == len(hierarchy):
                assert derived_from is quasardb.Error

        # Classes derived
        elif isinstance(xs, list):
            if depth == len(hierarchy) - 1:
                return [walk_and_check(x, depth + 1) for x in xs]

            elif isinstance(xs[0], tuple) and isinstance(xs[0][0], type) and str(xs[0][0]) == "<class 'AssertionError'>":
                ## HAXXX -- skip the tree of assertion exceptions, because we don't care about this
                pass

            else:
                walk_and_check(xs[0], depth + 1)
                walk_and_check(xs[1], depth + 1)

    walk_and_check(cur)



# pylint: disable=missing-module-docstring,missing-function-docstring,not-an-iterable,invalid-name

from datetime import timedelta
import logging
import numpy as np
import numpy.ma as ma
import pandas as pd
import pytest
import conftest

import quasardb
import quasardb.numpy as qdbnp
from utils import assert_indexed_arrays_equal, assert_ma_equal

logger = logging.getLogger("test-numpy")

def _unicode_to_object_array(xs):
    assert ma.isMA(xs)
    assert xs.dtype.kind == 'U'

    logger.debug("converting unicode to object array for proper testing / comparison")

    mask = xs.mask
    data = np.array([str(x) for x in xs.data],
                    dtype=np.dtype('O'))

    return ma.masked_array(data=data, mask=mask)

######
#
# Array tests
#
###
#
# The 'array' functions operate on just a single array. They use the column-oriented
# APIs, `qdb_ts_*_insert` under the hood.
#
######

@conftest.override_cdtypes('native')
def test_array_read_write_native_dtypes(array_with_index_and_table):
    """
     * qdbnp.write_array()
     * => qdb_ts_*_insert
     * => no conversion
    """
    (ctype, dtype, data, index, table) = array_with_index_and_table

    col = table.column_id_by_index(0)
    qdbnp.write_array(data, index, table, column=col, dtype=dtype, infer_types=False)

    res = qdbnp.read_array(table, col)

    assert_indexed_arrays_equal((index, data), res)


@conftest.override_cdtypes('inferrable')
def test_array_read_write_inferrable_dtypes(array_with_index_and_table):
    """
     * qdbnp.write_array()
     * => qdb_ts_*_insert
     * => conversion in python
    """
    (ctype, dtype, data, index, table) = array_with_index_and_table

    col = table.column_id_by_index(0)
    qdbnp.write_array(data, index, table, column=col, infer_types=True)

    res = qdbnp.read_array(table, col)
    assert_indexed_arrays_equal((index, data), res)


@conftest.override_cdtypes('native')
def test_arrays_read_write_native_dtypes(array_with_index_and_table, qdbd_connection):
    """
     * qdbnp.write_arrays()
     * => pinned writer
     * => no conversion
    """
    (ctype, dtype, data, index, table) = array_with_index_and_table

    col = table.column_id_by_index(0)
    qdbnp.write_arrays([data], qdbd_connection, table, index=index, dtype=dtype, infer_types=False, truncate=True)

    res = qdbnp.read_array(table, col)

    assert_indexed_arrays_equal((index, data), res)


@conftest.override_cdtypes('inferrable')
def test_arrays_read_write_inferrable_dtypes(array_with_index_and_table, qdbd_connection):
    """
     * qdbnp.write_arrays()
     * => pinned writer
     * => conversion in python
    """

    (ctype, dtype, data, index, table) = array_with_index_and_table

    col = table.column_id_by_index(0)
    qdbnp.write_arrays([data], qdbd_connection, table, index=index, infer_types=True)

    res = qdbnp.read_array(table, col)
    assert_indexed_arrays_equal((index, data), res)

@conftest.override_cdtypes('native')
def test_arrays_read_write_data_as_dict(array_with_index_and_table, qdbd_connection):
    """
     * qdbnp.write_arrays()
     * => pinned writer
     * => no conversion
    """
    (ctype, dtype, data, index, table) = array_with_index_and_table

    col = table.column_id_by_index(0)
    qdbnp.write_arrays({col: data}, qdbd_connection, table, index=index, dtype=dtype, infer_types=False, truncate=True)

    res = qdbnp.read_array(table, col)

    assert_indexed_arrays_equal((index, data), res)


######
#
# Arrays tests
#
###
#
# The 'arrays' functions operate on a collection of arrays, and use the 'exp' writer
# under the hood.
#
# They feature more advanced functionality, such as dropping duplicates and different
# insertion modes.
#
######

@conftest.override_sparsify('none')
def test_arrays_deduplicate(arrays_with_index_and_table, deduplication_mode, qdbd_connection):
    """
     * qdbnp.write_arrays()
     * => pinned writer
     * => deduplicate=True or a list
    """
    (ctype, dtype, arrays, index, table) = arrays_with_index_and_table

    if ctype == quasardb.ColumnType.String:
        print("SKIP -- PROBLEM IN UNICODE STRING COMPARISONS")
        print("See also: sc-11283/unicode-string-comparison-issue-with-masked-arrays-in-python-test-cases")
        return

    assert len(arrays) > 1
    data1 = arrays[0]
    data2 = arrays[1]

    col = table.column_id_by_index(0)

    # Validation:
    #
    # 1. write once, no deduplicate
    # 2. read intermediate data
    # 3. write again, with deduplicate
    # 4. read data again
    #
    # data from step 2 and 4 should be identical, as the write in step 3 should
    # drop all duplicates

    def _do_write_read(xs):
        """
        Utility function that makes it easier to insert data for this test, as they will all use
        the same column/table/index/etc.
        """
        qdbnp.write_arrays([xs],
                           qdbd_connection,
                           table,
                           deduplicate='$timestamp',
                           deduplication_mode=deduplication_mode,
                           index=index,
                           dtype=dtype,
                           infer_types=False)
        return qdbnp.read_array(table, col)

    res1 = _do_write_read(data1)
    assert_indexed_arrays_equal((index, data1), res1)

    # Regardless of deduplication_mode, since we're reinserting the same data, we always expect the results to be
    # identical.

    res2 = _do_write_read(data1)
    assert_indexed_arrays_equal((index, data1), res2)


    # Now, we're going to be inserting *different* data. Depending on the deduplication mode, the results will differ.
    res3 = _do_write_read(data2)

    if deduplication_mode == 'drop':
        # If we drop when deduplicating, and only deduplicate on $timestamp, then we expect
        # all new data to be dropped, because the entire index (i.e. all $timestamp) conflicts.
        assert_indexed_arrays_equal((index, data1), res3)

    else:
        assert deduplication_mode == 'upsert'
        # In this case, we expect all existing data to be replaced with the new data
        # assert_indexed_arrays_equal(res3, (index, data2))
        assert_indexed_arrays_equal((index, data2), res3)



######
#
# Miscellaneous tests
#
###
#
# Various tests which are important but don't belong anywhere else.
#
######

@conftest.override_cdtypes(np.dtype('unicode'))
def test_string_array_returns_unicode(array_with_index_and_table, qdbd_connection):
    """
    Validates that our C++ backend encodes unicode variable width arrays correctly.
    This yields significantly better performance than using objects, especially if
    you're not inspecting all of them.
    """
    (ctype, dtype, data, index, table) = array_with_index_and_table

    assert dtype == np.dtype('unicode')

    col = table.column_id_by_index(0)
    qdbnp.write_arrays([data], qdbd_connection, table, index=index)

    (idx, xs) = qdbnp.read_array(table, col)
    assert qdbnp.dtypes_equal(xs.dtype, np.dtype('unicode'))

######
#
# Query tests
#
###
#
# Validation of numpy query component. It's not as much the queries themselves we
# want to validate, but rather how our wrapper behaves: the validations it checks
# and errors it raises, its ability to turn the data in a certain shape, etc.
#
######

def _prepare_query_test(array_with_index_and_table, qdbd_connection):
    (ctype, dtype, data, index, table) = array_with_index_and_table

    col = table.column_id_by_index(0)
    qdbnp.write_arrays([data], qdbd_connection, table, index=index)
    q = 'SELECT $timestamp, {} FROM "{}"'.format(col, table.get_name())

    return (data, index, col, q)


def test_query_valid_results(array_with_index_and_table, qdbd_connection):
    (data, index, col, q) = _prepare_query_test(array_with_index_and_table, qdbd_connection)

    (idx, res) = qdbnp.query(qdbd_connection, q, index='$timestamp')
    assert_indexed_arrays_equal((index, data), (idx, res[0]))


def test_query_unknown_index(array_with_index_and_table, qdbd_connection):
    (data, index, col, q) = _prepare_query_test(array_with_index_and_table, qdbd_connection)

    with pytest.raises(KeyError):
        qdbnp.query(qdbd_connection, q, index='fooasbsdaog')


def test_query_iota_index(array_with_index_and_table, qdbd_connection):
    (data, index, col, q) = _prepare_query_test(array_with_index_and_table, qdbd_connection)

    # Simple test which just ensures when no explicit index is provided,
    # we get numbered "iota"-range back instead.
    (idx, res) = qdbnp.query(qdbd_connection, q)

    # A range of 1,2,3,4 etc will have a diff of exactly 1 for each item
    idx_ = np.diff(idx)
    assert np.all(idx_ == 1)


@conftest.override_sparsify('partial')
def test_query_masked_index(array_with_index_and_table, qdbd_connection):
    (data, index, col, q) = _prepare_query_test(array_with_index_and_table, qdbd_connection)

    with pytest.raises(ValueError):
        # We know we always have a partially masked array
        qdbnp.query(qdbd_connection, q, index=col)


def test_query_empty_result(array_with_index_and_table, qdbd_connection):
    (ctype, dtype, data, index, table) = array_with_index_and_table

    q = 'SELECT * FROM "{}"'.format(table.get_name())
    (idx, res) = qdbnp.query(qdbd_connection, q)

    assert len(idx) == 0
    assert len(res) == len(idx)


@conftest.override_sparsify('none')
@conftest.override_cdtypes([np.dtype('int64'),
                            np.dtype('float64')])
def test_query_insert(array_with_index_and_table, qdbd_connection):
    (ctype, dtype, data, index, table) = array_with_index_and_table

    col = table.column_id_by_index(0)
    val = data[0]

    q = 'INSERT INTO "{}" ($timestamp, {}) VALUES (NOW(), {})'.format(table.get_name(), col, val)
    (idx, res) = qdbnp.query(qdbd_connection, q)

    assert len(idx) == 0
    assert len(res) == len(idx)


def test_regression_sc10919_sc10933(qdbd_connection, table_name, start_date, row_count):
    t = qdbd_connection.table(table_name)

    # Specific regresion test used in Python user guide / API documentation
    cols = [quasardb.ColumnInfo(quasardb.ColumnType.Double, "open"),
            quasardb.ColumnInfo(quasardb.ColumnType.Double, "close"),
            quasardb.ColumnInfo(quasardb.ColumnType.Double, "high"),
            quasardb.ColumnInfo(quasardb.ColumnType.Double, "low"),
            quasardb.ColumnInfo(quasardb.ColumnType.Int64, "volume")]

    t.create(cols)

    idx = np.array([start_date + np.timedelta64(i, 's')
                    for i in range(row_count)]).astype('datetime64[ns]')

    # Note, partially sparse data just to increase some test surface
    data = {'open': np.random.uniform(100, 200, row_count),
            'close': np.random.uniform(100, 200, row_count),
            'volume': np.random.randint(10000, 20000, row_count)}

    qdbnp.write_arrays(data, qdbd_connection, t, index=idx, infer_types=False)

    # Retrieve just a single column, so that we know for sure we didn't
    # accidentally insert all null values.
    q = 'SELECT open FROM "{}"'.format(table_name)
    (idx, (res)) = qdbnp.query(qdbd_connection, q)



def test_regression_sc11333(qdbd_connection, table_name, start_date, row_count):
    """
    Ensures that we can provide data as numpy arrays as well as regular lists.
    """
    t = qdbd_connection.table(table_name)

    # Specific regresion test used in Python user guide / API documentation
    cols = [quasardb.ColumnInfo(quasardb.ColumnType.Double, "open"),
            quasardb.ColumnInfo(quasardb.ColumnType.Double, "close"),
            quasardb.ColumnInfo(quasardb.ColumnType.Int64, "volume")]
    t.create(cols)


    idx = np.array([start_date + np.timedelta64(i, 's')
                    for i in range(row_count)]).astype('datetime64[ns]')

    # Note, partially sparse data just to increase some test surface
    data = {'open': np.random.uniform(100, 200, row_count),
            'close': np.random.uniform(100, 200, row_count),
            'volume': np.random.randint(10000, 20000, row_count)}

    df = pd.DataFrame(data, index=idx)
    data = df.to_numpy()

    # Degenerate use case: by default, df.to_numpy() pivots the data incorrectly, in which case we
    # expect an error to be raised.
    assert len(data) == row_count
    assert len(data[0]) == len(cols)

    with pytest.raises(qdbnp.InvalidDataCardinalityError):
        qdbnp.write_arrays(data, qdbd_connection, t, index=idx, infer_types=False)

    data = data.transpose()
    qdbnp.write_arrays(data, qdbd_connection, t, index=idx, infer_types=True)

    cols = ['open', 'close', 'volume']
    for i in range(len(cols)):
        col = cols[i]

        res = qdbnp.read_array(t, col)
        assert_indexed_arrays_equal((idx, data[i]), res)



# pylint: disable=C0103,C0111,C0302,W0212
import datetime
import pytest
import numpy as np
import quasardb
import test_table as tslib


def test_timestamp_get_ranges__when_timeseries_is_empty(table, intervals):
    results = table.timestamp_get_ranges(tslib._ts_col_name(table), intervals)
    assert len(results) == 2
    assert len(results[0]) == 0
    assert len(results[1]) == 0


def test_timestamp_erase_ranges__when_timeseries_is_empty(table, intervals):
    erased_count = table.erase_ranges(tslib._ts_col_name(table), intervals)
    assert erased_count == 0


def test_timestamp_get_ranges(table, intervals):
    start_time = tslib._start_time(intervals)
    column_name = tslib._ts_col_name(table)

    inserted_timestamp_data = tslib._generate_timestamp_ts(
        start_time, start_time, 1000)
    table.timestamp_insert(column_name,
                           inserted_timestamp_data[0],
                           inserted_timestamp_data[1])

    results = table.timestamp_get_ranges(column_name, [(
        start_time, start_time + np.timedelta64(10, 's'))])

    tslib._check_ts_results(results, inserted_timestamp_data, 10)

    results = table.timestamp_get_ranges(column_name,
                                         [(start_time,
                                           start_time + np.timedelta64(10, 's')),
                                          (start_time + np.timedelta64(10, 's'),
                                             start_time + np.timedelta64(20, 's'))])

    tslib._check_ts_results(results, inserted_timestamp_data, 20)

    # Everything
    results = table.timestamp_get_ranges(column_name)
    tslib._check_ts_results(results, inserted_timestamp_data, 1000)

    # empty result
    out_of_time = start_time + np.timedelta64(10, 'h')
    results = table.timestamp_get_ranges(
        column_name, [(out_of_time, out_of_time + np.timedelta64(10, 's'))])
    assert len(results) == 2
    assert len(results[0]) == 0
    assert len(results[1]) == 0

    # error: column doesn't exist
    with pytest.raises(quasardb.Error):
        table.timestamp_get_ranges(
            "lolilol", [(start_time, start_time + np.timedelta64(10, 's'))])

    with pytest.raises(quasardb.Error):
        table.timestamp_insert(
            "lolilol",
            inserted_timestamp_data[0],
            inserted_timestamp_data[1])

    with pytest.raises(quasardb.IncompatibleTypeError):
        table.blob_get_ranges(
            column_name, [(start_time, start_time + np.timedelta64(10, 's'))])

    with pytest.raises(quasardb.IncompatibleTypeError):
        table.blob_insert(
            column_name,
            inserted_timestamp_data[0],
            inserted_timestamp_data[1])

    # These are a bit more tricky, because under the hood, a np.datetime64 is
    # represented as a int64. So verifying these is very important!
    with pytest.raises(quasardb.IncompatibleTypeError):
        table.int64_get_ranges(
            column_name, [(start_time, start_time + np.timedelta64(10, 's'))])

    with pytest.raises(quasardb.IncompatibleTypeError):
        table.int64_insert(
            column_name,
            inserted_timestamp_data[0],
            inserted_timestamp_data[1])


def test_timestamp_erase_ranges(table, intervals):
    start_time = tslib._start_time(intervals)
    column_name = tslib._ts_col_name(table)

    inserted_timestamp_data = tslib._generate_timestamp_ts(
        start_time, start_time, 1000)
    table.timestamp_insert(
        column_name,
        inserted_timestamp_data[0],
        inserted_timestamp_data[1])

    results = table.timestamp_get_ranges(column_name, [(
        start_time, start_time + np.timedelta64(10, 's'))])

    erased_count = table.erase_ranges(column_name, [(
        start_time, start_time + np.timedelta64(10, 's'))])

    assert erased_count == len(results[0])

    erased_count = table.erase_ranges(column_name, [(
        start_time, start_time + np.timedelta64(10, 's'))])

    assert erased_count == 0

    results = table.timestamp_get_ranges(column_name, [(
        start_time, start_time + np.timedelta64(10, 's'))])

    assert len(results[0]) == 0



# pylint: disable=C0103,C0111,C0302,W0212
from builtins import range as xrange  # pylint: disable=W0622
import pytest
import quasardb


def test_put(blob_entry, random_string):
    blob_entry.put(random_string)


def test_put_throws_exception_when_called_twice(blob_entry, random_blob):
    blob_entry.put(random_blob)
    with pytest.raises(quasardb.AliasAlreadyExistsError):
        blob_entry.put(random_blob)


def test_get(blob_entry, random_blob):
    blob_entry.put(random_blob)

    got = blob_entry.get()

    assert random_blob == got


def test_remove(blob_entry, random_blob):
    blob_entry.put(random_blob)
    blob_entry.remove()

    with pytest.raises(quasardb.AliasNotFoundError):
        blob_entry.get()


def test_remove_throws_exception_when_called_twice(blob_entry, random_blob):
    blob_entry.put(random_blob)
    blob_entry.remove()

    with pytest.raises(quasardb.AliasNotFoundError):
        blob_entry.remove()


def test_update(blob_entry, random_blob):
    blob_entry.update(random_blob)
    got = blob_entry.get()
    assert random_blob == got

    new_entry_content = b"It's the new style"
    blob_entry.update(new_entry_content)
    got = blob_entry.get()
    assert got == new_entry_content

    blob_entry.remove()

    new_entry_content = b"I'm never in training, my voice is not straining"
    blob_entry.update(new_entry_content)
    got = blob_entry.get()

    assert got == new_entry_content


def test_get_and_update(blob_entry, random_blob):
    blob_entry.put(random_blob)
    got = blob_entry.get()
    assert random_blob == got

    entry_new_content = b"new stuff"
    blob_entry.update(entry_new_content)
    got = blob_entry.get()

    assert entry_new_content == got

    got = blob_entry.get()
    assert entry_new_content == got
    blob_entry.remove()


def test_get_and_remove(blob_entry, random_blob):
    blob_entry.put(random_blob)

    got = blob_entry.get_and_remove()
    assert random_blob == got

    with pytest.raises(quasardb.AliasNotFoundError):
        blob_entry.get()


def test_remove_if(blob_entry, random_blob):
    blob_entry.put(random_blob)
    got = blob_entry.get()

    assert random_blob == got

    with pytest.raises(quasardb.Error):
        blob_entry.remove_if(random_blob + b'a')

    got = blob_entry.get()
    assert random_blob == got

    blob_entry.remove_if(random_blob)

    with pytest.raises(quasardb.AliasNotFoundError):
        blob_entry.get()


def test_compare_and_swap(blob_entry, random_blob):

    blob_entry.put(random_blob)
    got = blob_entry.get()

    assert random_blob == got

    entry_new_content = b"new stuff"
    got = blob_entry.compare_and_swap(entry_new_content, entry_new_content)

    assert random_blob == got

    got = blob_entry.compare_and_swap(entry_new_content, random_blob)
    assert len(got) == 0

    got = blob_entry.get()
    assert entry_new_content == got

    blob_entry.remove()



import copy
import quasardb
import numpy as np
import numpy.ma as ma

__all__ = []

def _ensure_ctype(self, idx, ctype):
    assert 'table' in self._legacy_state
    infos = self._legacy_state['table'].list_columns()
    cinfo = infos[idx]

    ctype_data = copy.copy(ctype)
    ctype_column = copy.copy(cinfo.type)

    if ctype_data == quasardb.ColumnType.Symbol:
        ctype_data = quasardb.ColumnType.String

    if ctype_column == quasardb.ColumnType.Symbol:
        ctype_column = quasardb.ColumnType.String

    if not ctype_data == ctype_column:
        raise quasardb.IncompatibleTypeError()


def _legacy_next_row(self, table):
    if 'pending' not in self._legacy_state:
        self._legacy_state['pending'] = []

    if 'table' not in self._legacy_state:
        self._legacy_state['table'] = table

    self._legacy_state['pending'].append({'by_index': {}})

    # Return reference to the row inside the buffer
    return self._legacy_state['pending'][-1]


def _legacy_current_row(self):
    return self._legacy_state['pending'][-1]


def _legacy_start_row(self, table, x):
    row = _legacy_next_row(self, table)
    assert '$timestamp' not in row
    row['$timestamp'] = x


def _legacy_set_double(self, idx, x):
    _ensure_ctype(self, idx, quasardb.ColumnType.Double)
    assert isinstance(x, float)
    assert idx not in _legacy_current_row(self)['by_index']
    _legacy_current_row(self)['by_index'][idx] = x


def _legacy_set_int64(self, idx, x):
    _ensure_ctype(self, idx, quasardb.ColumnType.Int64)
    assert isinstance(x, int)
    assert idx not in _legacy_current_row(self)['by_index']
    _legacy_current_row(self)['by_index'][idx] = x


def _legacy_set_timestamp(self, idx, x):
    _ensure_ctype(self, idx, quasardb.ColumnType.Timestamp)
    assert idx not in _legacy_current_row(self)['by_index']
    _legacy_current_row(self)['by_index'][idx] = x


def _legacy_set_string(self, idx, x):
    _ensure_ctype(self, idx, quasardb.ColumnType.String)
    assert isinstance(x, str)
    assert idx not in _legacy_current_row(self)['by_index']

    _legacy_current_row(self)['by_index'][idx] = x


def _legacy_set_blob(self, idx, x):
    _ensure_ctype(self, idx, quasardb.ColumnType.Blob)
    assert isinstance(x, bytes)
    assert idx not in _legacy_current_row(self)['by_index']

    _legacy_current_row(self)['by_index'][idx] = x


def _legacy_push(self):
    if 'pending' not in self._legacy_state:
        # Extremely likely default case, no "old" rows
        return

    assert 'table' in self._legacy_state
    table = self._legacy_state['table']

    # Some useful constants
    dtype_by_ctype = {quasardb.ColumnType.Double: np.dtype('float64'),
                      quasardb.ColumnType.Int64: np.dtype('int64'),
                      quasardb.ColumnType.Timestamp: np.dtype('datetime64[ns]'),
                      quasardb.ColumnType.String: np.dtype('unicode'),
                      quasardb.ColumnType.Symbol: np.dtype('unicode'),
                      quasardb.ColumnType.Blob: np.dtype('bytes')
                    }

    ctype_by_idx = {}
    cinfos = table.list_columns()
    for i in range(len(cinfos)):
        ctype_by_idx[i] = cinfos[i].type

    all_idx = set(ctype_by_idx.keys())

    # Prepare data structure
    pivoted = {'$timestamp': [],
               'by_index': {}}
    for i in all_idx:
        pivoted['by_index'][i] = []

    # Do the actual pivot
    for row in self._legacy_state['pending']:
        assert '$timestamp' in row
        assert 'by_index' in row

        pivoted['$timestamp'].append(row['$timestamp'])

        for idx in pivoted['by_index'].keys():
            val = row['by_index'].get(idx, None)
            pivoted['by_index'][idx].append(val)

    # Validation / verification, not strictly necessary. Effectively
    # ensures that we have the exact same amount of values for every
    # column
    for xs in pivoted['by_index'].values():
        assert len(xs) == len(pivoted['$timestamp'])

    column_data = []

    for idx,xs in pivoted['by_index'].items():
        ctype = ctype_by_idx[idx]
        dtype = dtype_by_ctype[ctype]

        # None-mask works, because everything inside the list are just regular ojbects

        mask = [x is None for x in xs]

        xs_ = []
        if all(mask):
            xs_ = ma.masked_all(len(xs),
                                dtype=dtype)
        else:
            xs_ = ma.masked_array(data=np.array(xs, dtype), mask=mask)

        assert len(xs_) == len(pivoted['$timestamp'])

        column_data.append(xs_)


    push_data = quasardb.WriterData()
    index = np.array(pivoted['$timestamp'], np.dtype('datetime64[ns]'))

    push_data.append(table, index, column_data)


    self._legacy_state = {}
    return push_data


def _wrap_fn(old_fn, replace_fn):

    def wrapped(self, *args, **kwargs):
        data = replace_fn(self)
        if data:
            return old_fn(self, data, *args, **kwargs)
        else:
            return old_fn(self, *args, **kwargs)

    return wrapped


def extend_writer(x):
    """
    Extends the writer with the "old", batch inserter API. This is purely
    a backwards compatibility layer, and we want to avoid having to maintain that
    in C++ with few benefits.
    """

    x.start_row      = _legacy_start_row
    x.set_double     = _legacy_set_double
    x.set_int64      = _legacy_set_int64
    x.set_string     = _legacy_set_string
    x.set_blob       = _legacy_set_blob
    x.set_timestamp  = _legacy_set_timestamp

    x.push           = _wrap_fn(x.push, _legacy_push)
    x.push_fast      = _wrap_fn(x.push_fast, _legacy_push)
    x.push_async     = _wrap_fn(x.push_async, _legacy_push)
    x.push_truncate  = _wrap_fn(x.push_truncate, _legacy_push)



# pylint: disable=C0103,C0111,C0302,W0212
from builtins import range as xrange  # pylint: disable=W0622
import pytest
import quasardb
import datetime

def test_put(timestamp_entry, random_timestamp):
    timestamp_entry.put(random_timestamp)


def test_get(timestamp_entry, random_timestamp):
    timestamp_entry.put(random_timestamp)
    got = timestamp_entry.get()

    assert random_timestamp == got


def test_update(timestamp_entry, random_timestamp):
    timestamp_entry.put(random_timestamp)
    got = timestamp_entry.get()

    assert random_timestamp == got


    new_timestamp = random_timestamp + datetime.timedelta(days=42)

    timestamp_entry.update(new_timestamp)
    got = timestamp_entry.get()

    assert new_timestamp == got



# pylint: disable=C0103,C0111,C0302,W0212
from builtins import range as xrange, int as long  # pylint: disable=W0622
from functools import reduce  # pylint: disable=W0622
import datetime
import test_table as tslib
from time import sleep

import pytest
import quasardb
import numpy as np


def _generate_data(count, start=np.datetime64('2017-01-01', 'ns')):
    integers = np.random.randint(-100, 100, count)
    timestamps = tslib._generate_dates(
        start + np.timedelta64('1', 'D'), count)

    return (integers, timestamps)

def test_incorrect_type_double(qdbd_connection, table):
    writer = qdbd_connection.writer()
    writer.start_row(table, np.datetime64('2020-01-01T00:00:00', 'ns'))
    for idx in range(6):
        if idx == 0:
            continue
        with pytest.raises(quasardb.IncompatibleTypeError):
            writer.set_double(idx, 1.1)


def test_successful_type_double(qdbd_connection, table):
    timestamp = np.datetime64('2020-01-01T00:00:00', 'ns')
    value = 1.1

    writer = qdbd_connection.writer()
    writer.start_row(table, timestamp)
    writer.set_double(0, value)
    writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_double" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 1
    assert res[0]['$timestamp'] == timestamp
    assert res[0]['the_double'] == value


def test_incorrect_type_blob(qdbd_connection, table):
    writer = qdbd_connection.writer()
    writer.start_row(table, np.datetime64('2020-01-01T00:00:00', 'ns'))
    with pytest.raises(quasardb.IncompatibleTypeError):
        writer.set_int64(1, 1234)

def test_successful_type_blob(qdbd_connection, table):
    timestamp = np.datetime64('2020-01-01T00:00:00', 'ns')
    value = b'aaa'

    writer = qdbd_connection.writer()
    writer.start_row(table, timestamp)
    writer.set_blob(1, value)
    writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_blob" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 1
    assert res[0]['$timestamp'] == timestamp
    assert res[0]['the_blob'] == value


def test_incorrect_type_string(qdbd_connection, table):
    writer = qdbd_connection.writer()
    writer.start_row(table, np.datetime64('2020-01-01T00:00:00', 'ns'))
    with pytest.raises(quasardb.IncompatibleTypeError):
        writer.set_int64(2, 1234)

def test_successful_type_string(qdbd_connection, table):
    timestamp = np.datetime64('2020-01-01T00:00:00', 'ns')
    value = 'aaa'

    writer = qdbd_connection.writer()
    writer.start_row(table, timestamp)
    writer.set_string(2, value)
    writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_string" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 1
    assert res[0]['$timestamp'] == timestamp
    assert res[0]['the_string'] == value


def test_incorrect_type_int64(qdbd_connection, table):
    writer = qdbd_connection.writer()
    writer.start_row(table, np.datetime64('2020-01-01T00:00:00', 'ns'))
    for idx in range(6):
        if idx == 3:
            continue
        with pytest.raises(quasardb.IncompatibleTypeError):
            writer.set_int64(idx, 1)


def test_successful_type_int64(qdbd_connection, table):
    timestamp = np.datetime64('2020-01-01T00:00:00', 'ns')
    value = 1

    writer = qdbd_connection.writer()
    writer.start_row(table, timestamp)
    writer.set_int64(3, value)
    writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 1
    assert res[0]['$timestamp'] == timestamp
    assert res[0]['the_int64'] == value


def test_incorrect_type_timestamp(qdbd_connection, table):
    writer = qdbd_connection.writer()
    writer.start_row(table, np.datetime64('2020-01-01T00:00:00', 'ns'))
    for idx in range(6):
        if idx == 4:
            continue
        with pytest.raises(quasardb.IncompatibleTypeError):
            writer.set_timestamp(
                idx, np.datetime64(
                    '2020-01-01T00:00:00', 'ns'))


def test_successful_type_timestamp(qdbd_connection, table):
    timestamp = np.datetime64('2020-01-01T00:00:00', 'ns')
    value = np.datetime64('2020-01-01T00:00:00', 'ns')

    writer = qdbd_connection.writer()
    writer.start_row(table, timestamp)
    writer.set_timestamp(4, value)
    writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_ts" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 1
    assert res[0]['$timestamp'] == timestamp
    assert res[0]['the_ts'] == value


def test_incorrect_type_symbol(qdbd_connection, table):
    writer = qdbd_connection.writer()
    writer.start_row(table, np.datetime64('2020-01-01T00:00:00', 'ns'))
    with pytest.raises(quasardb.IncompatibleTypeError):
        writer.set_int64(5, 1234)


def test_successful_type_symbol(qdbd_connection, table):
    timestamp = np.datetime64('2020-01-01T00:00:00', 'ns')
    value = 'aaa'

    writer = qdbd_connection.writer()
    writer.start_row(table, timestamp)
    writer.set_string(5, value)
    writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_symbol" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 1
    assert res[0]['$timestamp'] == timestamp
    assert res[0]['the_symbol'] == value


def test_insert_data_ordered_single_shard(qdbd_connection, table):
    writer = qdbd_connection.writer()

    timestamps = [
        np.datetime64(
            '2020-01-01T00:00:00',
            'ns'),
        np.datetime64(
            '2020-01-01T00:00:01',
            'ns')]
    values = [0, 1]

    writer.start_row(table, timestamps[0])
    writer.set_int64(3, values[0])
    writer.start_row(table, timestamps[1])
    writer.set_int64(3, values[1])

    writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 2
    for idx, row in enumerate(res):
        assert row['$timestamp'] == timestamps[idx]
        assert row['the_int64'] == values[idx]


def test_insert_data_ordered_multiple_shards(qdbd_connection, table):
    writer = qdbd_connection.writer()

    timestamps = [
        np.datetime64(
            '2020-01-01',
            'ns'),
        np.datetime64(
            '2020-01-02',
            'ns')]
    values = [1, 2]

    writer.start_row(table, timestamps[0])
    writer.set_int64(3, values[0])
    writer.start_row(table, timestamps[1])
    writer.set_int64(3, values[1])

    writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 2
    for idx, row in enumerate(res):
        assert row['$timestamp'] == timestamps[idx]
        assert row['the_int64'] == values[idx]


def test_insert_data_unordered_single_shard(qdbd_connection, table):
    writer = qdbd_connection.writer()

    timestamps = [
        np.datetime64(
            '2020-01-01T00:00:01',
            'ns'),
        np.datetime64(
            '2020-01-01T00:00:00',
            'ns')]
    values = [1, 0]

    writer.start_row(table, timestamps[0])
    writer.set_int64(3, values[0])
    writer.start_row(table, timestamps[1])
    writer.set_int64(3, values[1])

    writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))

    assert res[0]['$timestamp'] == timestamps[1]
    assert res[0]['the_int64'] == values[1]
    assert res[1]['$timestamp'] == timestamps[0]
    assert res[1]['the_int64'] == values[0]


def test_insert_data_unordered_multiple_shards(qdbd_connection, table):
    writer = qdbd_connection.writer()

    timestamps = [
        np.datetime64(
            '2020-01-02',
            'ns'),
        np.datetime64(
            '2020-01-01',
            'ns')]
    values = [2, 1]

    writer.start_row(table, timestamps[0])
    writer.set_int64(3, values[0])
    writer.start_row(table, timestamps[1])
    writer.set_int64(3, values[1])

    writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))

    assert res[0]['$timestamp'] == timestamps[1]
    assert res[0]['the_int64'] == values[1]
    assert res[1]['$timestamp'] == timestamps[0]
    assert res[1]['the_int64'] == values[0]


def test_insert_data_ordered_single_shard_two_push(qdbd_connection, table):
    writer = qdbd_connection.writer()

    timestamps = [
        np.datetime64(
            '2020-01-01T00:00:00',
            'ns'),
        np.datetime64(
            '2020-01-01T00:00:01',
            'ns')]
    values = [0, 1]

    writer.start_row(table, timestamps[0])
    writer.set_int64(3, values[0])
    writer.push()

    writer.start_row(table, timestamps[1])
    writer.set_int64(3, values[1])
    writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 2
    for idx, row in enumerate(res):
        assert row['$timestamp'] == timestamps[idx]
        assert row['the_int64'] == values[idx]

def test_insert_data_ordered_multiple_shards_two_push(qdbd_connection, table):
    writer = qdbd_connection.writer()

    timestamps = [
        np.datetime64(
            '2020-01-01',
            'ns'),
        np.datetime64(
            '2020-01-02',
            'ns')]
    values = [1, 2]

    writer.start_row(table, timestamps[0])
    writer.set_int64(3, values[0])
    writer.push()

    writer.start_row(table, timestamps[1])
    writer.set_int64(3, values[1])
    writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 2
    for idx, row in enumerate(res):
        assert row['$timestamp'] == timestamps[idx]
        assert row['the_int64'] == values[idx]

# generative tests


def _row_insertion_method(
        writer,
        table,
        dates,
        doubles,
        blobs,
        strings,
        integers,
        timestamps,
        symbols):
    for i in range(len(dates)):
        writer.start_row(table, dates[i])
        writer.set_double(0, doubles[i])
        writer.set_blob(1, blobs[i])
        writer.set_string(2, strings[i])
        writer.set_int64(3, integers[i])
        writer.set_timestamp(4, timestamps[i])
        writer.set_string(5, symbols[i])


def _regular_push(writer):
    writer.push()


def _async_push(writer):
    writer.push_async()
    # Wait for push_async to complete
    # Ideally we could be able to get the proper flush interval
    sleep(15)


def _fast_push(writer):
    writer.push_fast()


def _generate_data(count, start=np.datetime64('2017-01-01', 'ns')):
    doubles = np.random.uniform(-100.0, 100.0, count)
    integers = np.random.randint(-100, 100, count)
    blobs = np.array(list(np.random.bytes(np.random.randint(8, 16))
                          for i in range(count)), 'O')
    strings = np.array([("content_" + str(item)) for item in range(count)])
    timestamps = tslib._generate_dates(
        start + np.timedelta64('1', 'D'), count)
    symbols = np.array([("sym_" + str(item)) for item in range(count)])

    return (doubles, integers, blobs, strings, timestamps, symbols)


def _set_batch_writer_data(writer, table, intervals, data, start=0):
    (doubles, integers, blobs, strings, timestamps, symbols) = data

    for i in range(start, len(intervals)):
        writer.start_row(table, intervals[i])
        writer.set_double(0, doubles[i])
        writer.set_blob(1, blobs[i])
        writer.set_string(2, strings[i])
        writer.set_int64(3, integers[i])
        writer.set_timestamp(4, timestamps[i])
        writer.set_string(5, symbols[i])


def _assert_results(table, intervals, data):
    (doubles, integers, blobs, strings, timestamps, symbols) = data

    whole_range = (intervals[0], intervals[-1:][0] + np.timedelta64(2, 's'))
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], doubles)

    results = table.blob_get_ranges(tslib._blob_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], blobs)

    results = table.string_get_ranges(
        tslib._string_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], strings)

    results = table.int64_get_ranges(
        tslib._int64_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], integers)

    results = table.timestamp_get_ranges(
        tslib._ts_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], timestamps)

    results = table.string_get_ranges(
        tslib._symbol_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], symbols)
